# -*- coding: utf-8 -*-
"""
Created on Sat Oct  5 10:31:10 2013

@author: mwojc
"""

import numpy as np
import subprocess
import tempfile
import os

gext = ['.brep', '.brp',
        '.iges', '.igs',
        '.step', '.stp']

mext = ['.mesh']  # Medit files


class Gmsh(object):
    __eltypes = {1: (1, 'line2'),
                 2: (2, 'tria3'),
                 3: (2, 'quad4'),
                 4: (3, 'tetr4'),
                 5: (3, 'hexa8'),
                 6: (3, 'pris6'),
                 7: (3, 'pyra5'),
                 8: (1, 'line3'),
                 9: (2, 'tria6'),
                 10: (2, 'quad9'),
                 11: (3, 'tetr10'),
                 12: (3, 'hexa27'),
                 13: (3, 'pris18'),
                 14: (3, 'pyra14'),
                 15: (0, 'point'),
                 16: (2, 'quad8')}
    __meshio_eltypes = {
                'vertex':       15,
                'line':          1,
                'triangle':      2,
                'quad':          3,
                'tetra':         4,
                'hexahedron':    5,
                'wedge':         6,
                'pyramid':       7,
                'line3':         8,
                'triangle6':     9,
                'quad9':        10,
                'tetra10':      11,
                'hexahedron27': 12,
                'prism18':      13,
                'pyramid14':    14,
                'line4':        26,
                'quad16':       36,
                }
    __dims = {0: 'point', 1: 'curve', 2: 'surface', 3: 'volume'}

    def __init__(self, geometry, groups=None, entities=None,
                 custom=os.linesep, verbose=True):
        self.__groups = {}
        self.groups = {} if groups is None else groups
        self.entities = {} if entities is None else entities
        self.custom = custom
#        if geometry is None:  # create temporary geo file
#            geometry = self.__write_tempfile('', suffix='.geo')
        if isinstance(geometry, str):  # filename or .geo code
            try:
                open(os.path.realpath(geometry), 'r')
            except IOError:  # .geo code is assumed
                geometry = self.__write_tempfile(geometry, suffix='.geo')
            geometry = os.path.realpath(geometry)
            name, ext = os.path.splitext(geometry)
            if ext in ['.geo', '.msh']:
                pass
            elif ext in gext:  # create temporary geo file
                txt = 'Merge "%s";' % geometry
                geometry = self.__write_tempfile(txt, suffix='.geo')
            elif ext in mext:  # Convert meshes to msh
                sb = subprocess
                try:
                    output = sb.check_output(['gmsh', '-3', geometry],
                                             stderr=sb.STDOUT)
                except sb.CalledProcessError as err:
                    # if verbose:
                        # print err.output
                    raise Exception("Unknown mesh type: %s" % ext)
                geometry = name + '.msh'
            else:
                raise ValueError("Unknown geometry type: %s" % ext)
#            except IOError:
#                pass
#        elif isinstance(geometry, (tuple, list)):  # .geo and .msh files
#            pass
#        else:  # Geometry or Pygmsh objects
#            from fempy.geometry import Geometry as FempyGeometry
#            geometries = (FempyGeometry,)
#            try:
#                from pygmsh.built_in import Geometry as PygmshGeometry
#                geometries = geometries + (PygmshGeometry, )
#                # geometries.append(PygmshGeometry)
#            except:
#                pass
#            if not isinstance(geometry, geometries):
#                raise ValueError("Unknown geometry type: %s" % geometry)
        self.geometry = geometry
        self.verbose = verbose
        self.options = GmshOptions()
        self.options.factory = None  # or 'built-in' or 'opencascade'
        self.options_append = True
        self.custom_append = True

    @property
    def gmsh(self):
        return self

    def add_custom_line(self, line):
        line = str(line).strip()
        if not line.endswith(';'):
            line = line + ';'
        if not line.endswith(os.linesep):
            line = line + os.linesep
        self.custom += line

    def domain(self, order=None, etype=None, verbose=False):
        # Set options
        if order:
            if order in [1, 2]:
                self.options.Mesh.ElementOrder = order
            else:
                raise ValueError('Not supported order: %s' % order)
        self.verbose = verbose
        # Create fempy domain
        coors, conecs, etypes, egroups = self.__parse_msh()
        from fempy.domain.domain import Domain, tools
        from fempy.domain.index import EIndex
        dim = max(conecs.keys())
        etype_user = etype
        etype = list(set(etypes[dim]))
        if len(etype) > 1:
            raise TypeError("Mixed element types not supported: %s" % etype)
        else:
            etype = etype[0]
        # Enforce user element type if given
        etype = etype_user if etype_user is not None else etype
        #
        conec = np.asarray(conecs[dim], dtype=int)
        conec = conec.reshape(conec.shape[-2:])  # for _parse_msh, _parse_msh_meshio compat
        lst = tools.unique_rows(conec, sort=True, strict=False)
        lst = np.nonzero(lst)[0]
        eidx = EIndex(conec, lst, coors)
#        # RENUMBER NODES - seems to not help...
#        dom = Domain(coors, conec, etype)
#        segments = dom.segments
#        segments = dom.eidx.nidx.g.lst[segments]
#        v = np.ones(len(segments), dtype=int)
#        from scipy import sparse
#        M = sparse.coo_matrix((v, segments.T), dtype=int).tocsr()
#        perm = sparse.csgraph.reverse_cuthill_mckee(M, symmetric_mode=True)
##        import pylab
##        M = M[perm]
##        M = M[:, perm]
##        pylab.spy(M, markersize=2)
#        coors = coors[perm]
#        iperm = np.zeros(len(perm), dtype=int)
#        iperm[perm] = np.arange(len(perm), dtype=int)
#        conec = iperm[conec]
#        for i in range(dim-1, -1, -1):
#            conecs[i] = iperm[np.asarray(conecs[i], dtype=np.int)]
#        # END RENUMBERING NODES
        dom = Domain(coors, eidx, etype)  # , egroups[dim])
        for key, gpos in egroups[dim].items():
            dom.egroups.set_group_by_conec(key, conec[gpos], glob=True)
        faces = dom.faces
        for i in range(dim-1, -1, -1):
            if i in egroups:
                for key, gpos in egroups[i].items():
                    fconec = np.asarray(conecs[i], dtype=np.int)
                    fconec = fconec.reshape(fconec.shape[-2:])[gpos] # compat
                    faces.egroups.set_group_by_conec(key, fconec, glob=True)
            faces = faces.faces
        dom.geometry = self.geometry
        return dom

    @property
    def geo(self):
        self.__groups.clear()
        g = self.geometry
        if isinstance(g, str):  # possibly .geo file
            name, ext = os.path.splitext(g)
            if ext == '.geo':
                string = open(g, 'r').read()
            else:  # msh file
                return ''
        elif isinstance(g, (tuple, list)):  # .geo and .msh file
            string = open(g[0], 'r').read()
        elif hasattr(g, 'get_code'):
            # Pygmsh only currently
            string = g.get_code()
        else:
            # Geometry object is assumed
            string = ''
            string += self.__points(g.points)
            string += self.__curves(g.points, g.curves)
            string += self.__surfaces(g.curves, g.surfaces)
            string += self.__physical_points(g.points)
            string += self.__physical_curves(g.curves)
            string += self.__physical_surfaces(g.surfaces)

        if self.custom_append:
            string += self.custom

        if self.options_append:
            string += self.options.options

        return string

    @property
    def msh(self):
        g = self.geometry
        if isinstance(g, str):  # possibly .msh file
            name, ext = os.path.splitext(g)
            if ext == '.msh':
                return open(g, 'r').read()
            else:
                pass  # assumed the string is a .geo file (read later)
        elif isinstance(g, (tuple, list)):  # .geo and .msh file
            return open(g[1], 'r').read()
        else:
            pass  # Geometry object is assumed

        gfn = self.__write_tempfile(self.geo)
        mfn = gfn+'.msh'
        try:
            output = subprocess.check_output(['gmsh', '-3', gfn],
                                             stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as err:
#            if self.verbose:
#                print err.output
                #print err.returncode
            raise Exception('Mesh cannot be generated because of errors.')
#        if self.verbose:
#            print output
        mfn_file = open(mfn, 'r')
        m = mfn_file.read()
        mfn_file.close()
        os.remove(gfn)
        os.remove(mfn)
        return m

    def save(self, fname):
        import os
        name, ext = os.path.splitext(fname)
        if ext == '.geo':
            f = open(fname, 'w')
            f.write(self.geo)
            f.close()
        elif ext == '.msh':
            f = open(fname, 'w')
            f.write(self.msh)
            f.close()
        else:
            fg = open(fname+'.geo', 'w')
            fm = open(fname+'.msh', 'w')
            fg.write(self.geo)
            fm.write(self.msh)
            fg.close()
            fm.close()

    def preview(self, default='geomsh'):
        if default == 'geomsh':
            mfn = self.__write_tempfile(self.msh)
            gfn = self.__write_tempfile(self.geo)
            out = subprocess.call(['gmsh', gfn, mfn], stderr=subprocess.STDOUT)
            os.remove(mfn)
            os.remove(gfn)
        elif default == 'geo':
            gfn = self.__write_tempfile(self.geo)
            out = subprocess.call(['gmsh', gfn], stderr=subprocess.STDOUT)
            os.remove(gfn)
        elif default == 'msh':
            mfn = self.__write_tempfile(self.msh)
            out = subprocess.call(['gmsh', mfn], stderr=subprocess.STDOUT)
            os.remove(mfn)
        else:
            out = None
        return out

    def __write_tempfile(self, txt, suffix=''):
        f = tempfile.NamedTemporaryFile(mode='w', suffix=suffix, delete=False)
        f.write(txt)
        name = f.name
        f.close()
        return name

    def __parse_msh(self):
        try:
            return self._parse_msh_meshio()
        except ImportError:
            pass
        # Read file
        msh = self.msh.split('\n')
        msh.reverse()
        line = msh.pop()
        while(line != '$Nodes'):        # Read until we find the nodes
            line = msh.pop()
        msh.pop()                       # Skip next line (number of nodes)
        coors = []
        line = msh.pop()
        while(line != '$EndNodes'):     # Read nodes
            coors += [list(map(float, line.split()[1:]))]
            line = msh.pop()
        while(line != '$Elements'):     # Read until we find elements
            line = msh.pop()
        msh.pop()                       # Skip next line (number of elements)
        elinfo = []
        line = msh.pop()
        while(line != '$EndElements'):  # Read elements
            elinfo += [list(map(int, line.split()[1:]))]
            line = msh.pop()
        # End read file
        # Coordinates - remove unnecessary dimensions (3D --> 2D mostly)
        from fempy.geometry import tools
        coors = tools.remove_dimensions(coors)
        # Parse elements

        def parse_line(el):
            dim, etype = self.__eltypes[el.pop(0)]
            ntags = el.pop(0)
            group = el.pop(0)
            entity = el.pop(0)
            for j in range(ntags-2):
                el.pop(0)  # skip other tags
            con = np.asarray(el) - 1
            return dim, etype, ntags, group, entity, con

        conecs = {}
        etypes = {}
        egroups = {}
        for i in range(len(elinfo)):
            el = elinfo.pop(0)
            dim, etype, ntags, group, entity, con = parse_line(el)
            if dim not in conecs:
                conecs[dim] = []
            if dim not in etypes:
                etypes[dim] = []
            if dim not in egroups:
                egroups[dim] = {}
            #####
            pi = len(conecs[dim])
            if pi == 0:
                conecs[dim].append(con)
                etypes[dim].append(etype)
            else:  # pi > 0
                pcon = conecs[dim][-1]
                if not np.array_equiv(pcon, con):  # if repeated elem
                    conecs[dim].append(con)
                    etypes[dim].append(etype)
                else:
                    pi -= 1
            # Get physical group
            try:
                group = self.__groups[dim][group-1]
            except:
                try:
                    group = self.groups[dim][group]
                except:
                    group = 'gmsh_physical_%s_%s' % (self.__dims[dim], group)
            # Add physical group
            try:
                egroups[dim][group].append(pi)
            except:
                egroups[dim][group] = [pi]
            # Get and add gmsh entity group
            try:
                entity = self.entities[dim][entity]
            except:
                entity = 'gmsh_%s_%s' % (self.__dims[dim], entity)
            try:
                if egroups[dim][entity][-1] != pi:
                    egroups[dim][entity].append(pi)
            except:
                egroups[dim][entity] = [pi]

        return coors, conecs, etypes, egroups

    def _parse_msh_meshio(self):
        from fempy.geometry import gmsh_io
        msh = self.msh
        mfn = self.__write_tempfile(msh, suffix='.msh')
        points, cells, point_data, cell_data, field_data = gmsh_io.read(mfn)
        os.remove(mfn)
        # Coordinates - remove unnecessary dimensions (3D --> 2D mostly)
        from fempy.geometry import tools
        coors = tools.remove_dimensions(points)
        conecs = {}
        etypes = {}
        egroups = {}
        for k, v in cells.items():
            n = self.__meshio_eltypes[k]
            dim, etype = self.__eltypes[n]
            if dim not in conecs:
                conecs[dim] = []
            if dim not in etypes:
                etypes[dim] = []
            if dim not in egroups:
                egroups[dim] = {}
            conecs[dim].append(v)
            etypes[dim].append(etype)
            # Now geometrical groups
            gdata = cell_data[k]['geometrical']
            entities = np.unique(gdata)
            for en in entities:
                try:
                    entity = self.entities[dim][en]
                except:
                    entity = 'gmsh_%s_%s' % (self.__dims[dim], en)
                egroups[dim][entity] = np.nonzero(gdata == en)[0]
            # Now physical groups
            pdata = cell_data[k]['physical']
            if np.allclose(pdata, 0):
                continue  # do not add 0 physical group
            groups = np.unique(pdata)
            for gn in groups:
                try:
                    group = self.__groups[dim][gn - 1]  # -1 necessary here!
                except:
                    try:
                        group = self.groups[dim][gn]
                    except:
                        group = 'gmsh_physical_%s_%s' % (self.__dims[dim], gn)
                egroups[dim][group] = np.nonzero(pdata == gn)[0]
        return coors, conecs, etypes, egroups

    def __points(self, points):
        string = ''
        for i, point in enumerate(points):
            coors = point.coors.tolist().__str__()[1:-1]
            # if not hasattr(point, 'elsize'):
            if point.elsize is None:
                string += 'Point(%i) = {%s};\n' % (i+1, coors)
            else:
                elsize = point.elsize * point.elsize_factor
                #if hasattr(point, 'elsize_factor'):
                #    elsize = elsize * point.elsize_factor
                string += 'Point(%i) = {%s, %s};\n' % (i+1, coors, elsize)
        return string

    def __physical_points(self, points):
        groups = []
        towrite = {}
        for i, point in enumerate(points):
            for group in point.groups:
                if group not in groups:
                    groups += [group]
                    ip = len(groups)
                    towrite[ip] = [i+1]
                else:
                    ip = groups.index(group) + 1
                    towrite[ip] += [i+1]
        self.__groups[0] = groups  # Save groups
        string = ''
        for ip in sorted(towrite.keys()):
            pts = str(towrite[ip])[1:-1]
            string += 'Physical Point(%i) = {%s};\n' % (ip, pts)
        return string

    def __curves(self, points, curves):
        string = ''
        periodic = ''
        for i, curve in enumerate(curves):
            cname = curve.__class__.__name__
            if cname == 'Line':
                start = points.index(curve.start) + 1
                end = points.index(curve.end) + 1
                string += 'Line(%i) = {%i, %i};\n' % (i+1, start, end)
            elif cname == 'CircularArc':
                start = points.index(curve.start) + 1
                center = points.index(curve.center) + 1
                end = points.index(curve.end) + 1
                string += 'Circle(%i) = {%i, %i, %i};\n' \
                          % (i+1, start, center, end)
            elif cname == 'EllipticalArc':
                start = points.index(curve.start) + 1
                center = points.index(curve.center) + 1
                major = points.index(curve.major) + 1
                end = points.index(curve.end) + 1
                string += 'Ellipse(%i) = {%i, %i, %i, %i};\n' \
                          % (i+1, start, center, major, end)
            elif cname == 'Spline':
                pts = []
                for p in curve.points:
                    pts += [points.index(p) + 1]
                spts = str(pts)[1:-1]
                string += 'Spline(%i) = {%s};\n' \
                          % (i+1, spts)
            else:
                raise TypeError('Unknown curve type')
            # Handle periodic slaves
            if curve.slaves is not None:
                for slave, sign in curve.slaves.items():
                    n = (curves.index(slave)+1) # * np.sign(sign)
                    ni = (i+1) * np.sign(sign)
                    periodic += 'Periodic Line {%i} = {%i};\n' % (n, ni)
        return string + periodic

    def __physical_curves(self, curves):
        groups = []
        #curves = self.geometry.curves
        towrite = {}
        for i, curve in enumerate(curves):
            for group in curve.groups:
                if group not in groups:
                    groups += [group]
                    ip = len(groups)
                    towrite[ip] = [(i+1) * curve.orientation]
                else:
                    ip = groups.index(group) + 1
                    towrite[ip] += [(i+1) * curve.orientation]
        string = ''
        self.__groups[1] = groups  # Save groups
        for ip in sorted(towrite.keys()):
            crv = str(towrite[ip])[1:-1]
            string += 'Physical Line(%i) = {%s};\n' % (ip, crv)
        return string

    def __surfaces(self, curves, surfaces):
        curve_loops = []
        lstring = ''
        sstring = ''
        for i, surface in enumerate(surfaces):
            # Line loops
            llst = []
            for loop in surface.loops:
                if loop not in curve_loops:
                    curve_loops += [loop]
                    icl = len(curve_loops)
                    clst = [(curves.index(c)+1)*o for c, o in zip(*loop)]
                    cstr = str(clst)[1:-1]
                    lstring += 'Line Loop(%i) = {%s};\n' % (icl, cstr)
                else:
                    icl = curve_loops.index(loop) + 1
                llst += [icl]
            # Surfaces
            lstr = str(llst)[1:-1]
            if surface.__class__.__name__ == 'PlaneSurface':
                sstring += 'Plane Surface(%i) = {%s};\n' % (i+1, lstr)
            else:
                raise TypeError('Unknown surface type')
        return lstring + sstring

    def __physical_surfaces(self, surfaces):
        groups = []
        towrite = {}
        for i, surface in enumerate(surfaces):
            for group in surface.groups:
                if group not in groups:
                    groups += [group]
                    ip = len(groups)
                    towrite[ip] = [i+1]
                else:
                    ip = groups.index(group) + 1
                    towrite[ip] += [i+1]
        string = ''
        self.__groups[2] = groups  # Save groups
        for ip in sorted(towrite.keys()):
            srf = str(towrite[ip])[1:-1]
            string += 'Physical Surface(%i) = {%s};\n' % (ip, srf)
        return string


class GmshOptions(object):
    """
    Container for Gmsh options
    """
    # General options
    class __General(object):
        # write General options here
        pass
    General = __General()

    # Print options
    class __Print(object):
        # write Print options here
        pass
    Print = __Print()

    # Geometry options
    class __Geometry(object):
        # write Geometry options here
        pass

        class __Color(object):
            # write Geometry.Color options here
            pass
        Color = __Color()
    Geometry = __Geometry()

    # Mesh options
    class __Mesh(object):
        # write Mesh options here
        #
        # Mesh coloring (0=by element type, 1=by elementary entity,
        # 2=by physical entity, 3=by partition)
        ColorCarousel = 2
        # Display faces of surface mesh?
        SurfaceFaces = 1
        # Use vesion 2 of msh file format
        MshFileVersion = 2

        class __Color(object):
            # write Mesh.Color options here
            pass
        Color = __Color()
    Mesh = __Mesh()

    # Solver options
    class __Solver(object):
        # write Solver options here
        pass
    Solver = __Solver()

    # PostProcessing options
    class __PostProcessing(object):
        # write PostPtocessing options here
        #
        # Delay (in seconds) between frames in automatic animation mode
        AnimationDelay = 0.25
    PostProcessing = __PostProcessing()

    # View options
    class __View(object):
        pass
    View = __View()

    @property
    def options(self):
        string = '\n// OPTIONS\n'
        for group in ['General', 'Print', 'Geometry', 'Mesh',
                      'Solver', 'PostProcessing', 'View']:
            string += self.__write_options(group)
        # Apply built-in factory
        if self.factory is not None:
            string += 'SetFactory("%s");\n' % self.factory
        return string

    def __get_options(self, obj):
        options = [attr for attr in dir(obj)
                   if not callable(getattr(obj, attr))
                   and not attr.startswith("__")]
        return options

    def __write_options(self, group):
        string = ''
        opts = getattr(self, group)
        for opt in self.__get_options(opts):
            if opt == 'Color':
                colors = getattr(opts, 'Color')
                for cname in self.__get_options(colors):
                    cval = getattr(colors, cname)
                    if cval is not None:
                        cval = '{' + str(cval)[1:-1] + '}'
                        string += '%s.Color.%s = %s;\n' % (group, cname, cval)
            else:
                val = getattr(opts, opt)
                if val is not None:
                    if isinstance(val, str):
                        val = '"' + str(val) + '"'
                    val = str(val)
                    string += '%s.%s = %s;\n' % (group, opt, val)
        return string


# Tests
if __name__ == '__main__':
    from fempy.testing import testmodule
    testmodule(__file__, '-v')



# -*- coding: utf-8 -*-
"""
Created on Fri Jun 27 21:30:46 2014

@author: mwojc
"""
from fempy.geometry.shapes import Rectangle
from fempy.models.elastic import PlaneStrain
from fempy.solvers import nonlin
import numpy as np


class MyModel(PlaneStrain):
    default_solver = nonlin.SimpleIterationSolver

    def setup(self):
        self.ev(20000, 0.3)
        self.rho[:] = 2.
        self.gravity[:] = [0., -9.81]
        self.p0 = -10.
        self.utop = -0.0005
        self.time0 = 0.
        self.time1 = 1.
        self.time = 0.

    def pressures_update(self):
        t = self.time
        p0 = self.p0
        p = np.interp(t,
                      [0, 0.25, 0.5,  1.],
                      [0, p0,   2*p0, -p0])
        self.p['left'] = p

    def displacement_update(self):
        t = self.time
        utop = self.utop
        u = np.interp(t,
                      [0, 0.5,  1.],
                      [0, utop, 0])
        self.u['top', 1] = u

    def bcs_update(self):
        # where we update bcs
        where = ('right', 0)
        # get current bcs, u and r
        bcs = self.bcs[where]
        u = self.u[where]
        r = self.res[where]
        # update bcs and u
        bcs[u > 1e-4] = True
        bcs[r > 0.] = False
        u[bcs] = 1e-4
        # assign bcs and u
        self.u[where] = u
        self.bcs[where] = bcs
        # assign other necessary bcs
        self.bcs['top', 1] = True
        self.bcs['bottom'] = True
        self.u['bottom'] = 0.


d = Rectangle(elsize=0.05).gmsh.domain(order=2)
m = MyModel(d)
m.solve(steps=50, verbose=True)
m.preview('all')



# -*- coding: utf-8 -*-
"""
Created on Tue Jun  3 12:35:53 2014

@author: mwojc
"""
from fempy.compat import builtins
import numpy as np
from fempy.fields.nfields import nScalar, nVector
from fempy.fields.gfields import gScalar, gVector, gTensor, gMatrix
from fempy.fields.algebra import normalize
from fempy.domain.elems import guess_etype
from fempy.domain.index import EIndex, EGroups, NGroups
import fempy.domain.tools as tools

# We will use cached property throughout domains
property = tools.cached_property


def Domain(coors, conec, etype=None, egroups=None, ngroups=None):
    """
    Factory of finite element method domains

    Arguments:
    coors   -   node coordinates, indices are node numbers, possibly more
                coordinates then needed (by conec) can be given
    conec   -   connectivity array from which nodes being used are determined,
                if this is an EIndex object than it is used directly (coors
                is ignored)
    etype   -   string with element name or element class/module
    egroups -   element groups dictionary (global)
    ngroups -   node groups dictionary (global)

    Returns:
    FEM domain, one of: Volume, Surface, Surface3D, Curve, Curve2D, Curve3D,
    Points
    """
    eidx = conec if isinstance(conec, EIndex) else EIndex(conec, nidx=coors)
    eidx.g.update(egroups)
    eidx.nidx.g.update(ngroups)
    etype = guess_etype(eidx.nidx.g.coors, eidx.g.conec, etype)
    dtype = guess_dtype(eidx.nidx.g.coors.shape[1], etype.ndime)
    return dtype(eidx, etype)


class BaseDomain(object):
    """
    Base FEM domain consisting in elements of the same type.
    """
    def __init__(self, eidx, etype=None):
        """
        Create FEM domain.

        eidx  - element index (EIndex object)
        etype - element type
        """
        if not isinstance(eidx, EIndex):  # assume eidx is coors and conec
            eidx = EIndex(eidx[1], nidx=eidx[0])
        self.eidx = eidx
        self.etype = guess_etype(eidx.nidx.g.coors, eidx.g.conec, etype)
        self.egroups = EGroups(self)
        self.ngroups = NGroups(self)
        self._faces = None
        self._boundary = None
        self.geometry = None

    @property
    def faces(self):
        if self._faces is None:
            self._faces = tools.create_faces(self)
        return self._faces

    @property
    def boundary(self):
        if self._boundary is None:
            f = self.faces
            if f is not None:
                self._boundary = f['boundary']
        return self._boundary

    @property
    def segments(self):
        if self.ndime == 0:
            return None
        elif self.ndime == 1:
            edges = self
        elif self.ndime == 2:
            edges = self.faces
        else:
            edges = self.faces.faces
        return edges.conec[:, edges.etype.segments].reshape(-1, 2)

    def __getitem__(self, key):
        self.faces  # creates boundary if possible and necessary
        subeidx = self.eidx.subindex(key)
        subdomain = self.__class__(subeidx, self.etype)
        # Handle faces
        # This cannot be as simple as tools.create_faces because also
        # for faces we want to have subdomian not the new faces domain
        if self.etype.btype is not None:
            # global conec for domain faces
            fconec0 = self.faces.eidx.conec_g
            # global conec for subdomain faces
            fconec1 = subdomain.eidx.conec_g[:, self.etype.faces]
            fconec1 = fconec1.reshape(-1, fconec1.shape[-1])
            # find which subdomain faces are in domain faces
            mask = tools.in2d(fconec0, fconec1)  # no sort!
            # be sure only one occurence of each face is taken
            mask2 = tools.unique_rows(fconec0[mask], strict=False)
            mask[mask] = mask2
            # create subindex
            subfidx = self.faces.eidx.subindex(mask)
            # now we will identify boundary elements
            bbool = tools.unique_rows(subfidx.conec, sort=True)
            # and we apply local boundary masks
            subfidx.set_local_mask('boundary', bbool)
            subfidx.set_local_mask('CUT', subfidx['boundary^BOUNDARY'])
            # TODO this below hack is needed to recover proper node indices
            subeidx.nidx.set_local_mask('boundary', subfidx.conec[bbool])
            subeidx.nidx.set_local_mask('CUT', subfidx.conec[subfidx['CUT']])
            # create and assign subfaces
            subfaces = self.faces.__class__(subfidx, self.faces.etype)
            subdomain._faces = subfaces
        return subdomain

    @property
    def nlist(self):
        return self.eidx.nidx.g.pos[self.eidx.nidx.lst]

    @property
    def ndime(self):
        return self.eidx.nidx.g.coors.shape[1]

    @property
    def nnode(self):
        return self.eidx.nidx.n

    @property
    def nedim(self):
        return self.etype.ndime

    @property
    def nenod(self):
        return self.etype.nnode

    @property
    def nelem(self):
        return self.eidx.n

    @property
    def ngauss(self):
        return self.etype.ngauss

    @property
    def coors(self):
        coors = self.eidx.nidx.g.coors
        idx = self.eidx.nidx.msk
        return nVector(self, coors[idx])

    @property
    def conec(self):
        return self.eidx.nidx.pos[self.eidx.conec_g]

    @property
    def ndeg(self):
        ndeg = np.bincount(self.conec.ravel())
        return nScalar(self, ndeg)

    @property
    def ecoors(self):
        coors = self.eidx.nidx.g.coors
        conec = self.eidx.conec_g
        return coors[conec]

    @property
    def ecoors_mean(self):
        ecoors = self.ecoors
        return np.mean(ecoors, axis=1)

    @property
    def gcoors(self):
        gcoors = np.einsum('eni, gn -> egi', self.ecoors, self.etype.gshp)
        return gVector(self, gcoors)

    @property
    def gjac(self):
        gjac = np.einsum('eni, gnj -> egij', self.ecoors, self.etype.gshpd)
        return gTensor(self, gjac)

    @property
    def uvw(self):
        return normalize(self.gjac, axis=-2)

    @property
    def gdet(self):
        detj = np.linalg.det(self.gjac)
        if np.any(detj <= 0.):
            raise ValueError("Determinant of jacobian <= 0")
        return gScalar(self, detj)

    @property
    def gvol(self):
        gvol = np.einsum('eg, g -> eg', self.gdet, self.etype.gweights)
        return gScalar(self, gvol)

    @property
    def evol(self):
        evol = self.gvol.sum(1)
        return evol

    @property
    def vol(self):
        return self.gvol.sum()

#    @property
#    def nvol(self):
#        from fempy.fields.operators import lform
#        nvol = nScalar(self)
#        nvol[:] = lform(gScalar(self, 1.))  # lform currently returns array
#        return nvol
#
#    @property
#    def nvolu(self):
#        nvol = np.zeros((self.nelem, self.etype.nnode))
#        nvol[:] = self.evol[:, None]/self.etype.nnode
#        F = nScalar(self)
#        np.add.at(F, self.eidx.conec_a, nvol)
#        return F
#
#    @property
#    def bvol(self):
#        from fempy.fields.operators import onboundary
#        return onboundary(self.nvol)
#
#    @property
#    def bgvol(self):
#        from fempy.fields.operators import ongauss
#        return ongauss(self.bvol)

    @property
    def gijac(self):
        gijac = np.linalg.inv(self.gjac)
        return gTensor(self, gijac)

    @property
    def egn(self):
        e = np.ones(len(self.conec))
        gshp = np.einsum('e, gn -> egn', e, self.etype.gshp)
        return gshp

    @property
    def egnk(self):
        gijac = self.gijac
        if self.etype.ndime < self.ndime:  # for structural elements
            d = self.etype.ndime
            gijac = gijac[..., :d, :]
        gshpd = np.einsum('egik, gni -> egnk', gijac, self.etype.gshpd)
        return gshpd

    def invalidate_caches(self):
        tools.invalidate_caches(self)
        # invalidate reccurently at boundary...
        if self.faces is not None:
            self.faces.invalidate_caches()
        if self.boundary is not None:
            self.boundary.invalidate_caches()

    def update(self, disp=None):
        if disp is not None:
            self.eidx.nidx.g.coors[self.eidx.nidx.lst] += disp
        self.invalidate_caches()

    def preview(self):
        from fempy.geometry import gmsh_io
        gmsh_io.preview_domain(self)

    # ### Axisymmetry specific code
    # TODO: the below attributes should not be visible in 3D
    @builtins.property
    def axis(self):
        if not hasattr(self, '_axis'):
            self._axis = None
        return self._axis

    @axis.setter
    def axis(self, ax):
        if self.ndime > 2 and ax is not None:
            raise AttributeError("Can't set 'axis' attribute for 3D domains")
        # If ax is None (back to plane surface)
        if ax is None:
            self._axis = None
            self.ngroups.pop('axis', None)  # removes 'axis' group if exists
            self.invalidate_caches()
            return
        # Otherwise (if 'auto' set minimum x coordinate).
        x = self.coors[..., 0]
        if ax in ['auto', 'minx']:
            ax = x.min()
        assert isinstance(ax, (np.float, np.int)), "Axis must be a number" \
                                                   " (x coordinate)"
        assert (x >= ax).all(), "All x coordinates must be greater than axis"
        self._axis = float(ax)
        # when axis is changed other things change
        self.invalidate_caches()
        # set the same axis to the faces and boundary
        if self.faces is not None:
            self.faces.axis = ax
        if self.boundary is not None:
            self.boundary.axis = ax
        # Set 'axis' ngroup
        rn = self.radiusn  # always > 0
        self.ngroups['axis'] = rn < np.spacing(rn)*2

    @property
    def radius(self):
        if self.axis is None:
            return None
        radius = self.gcoors[..., 0] - self.axis
        return gScalar(self, radius)
    radiusg = radius

    @property
    def radiusn(self):
        if self.axis is None:
            return None
        radiusn = self.coors[..., 0] - self.axis
        return nScalar(self, radiusn)


class Volume(BaseDomain):
    """
    3D volume
    """


class Surface(BaseDomain):
    """
    2D surface
    """
    def __getitem__(self, key):
        subdomain = BaseDomain.__getitem__(self, key)
        subdomain.thickness[:] = self.thickness[key]
        subdomain.axis = self.axis
        # TODO: how to properly inherit data in subdomain? Question relevant
        # also for Curve and Points
        return subdomain

    @property
    def _gvol(self):
        return super(Surface, self).gvol

    @property
    def thickness(self):
        # TODO: we should keep the boundary thickness always compatibile
        # superdomain concept is necessary here!
        if self.axis is not None:
            return 2*np.pi * self.radius
        if not hasattr(self, '_thickness'):
            self._thickness = gScalar(self, 1.)
        return self._thickness

    @builtins.property
    def gvol(self):
        return self._gvol * self.thickness


class Curve(BaseDomain):
    """
    1D line
    """
    def __getitem__(self, key):
        subdomain = BaseDomain.__getitem__(self, key)
        subdomain.cross[:] = self.cross[key]
        subdomain.axis = self.axis
        return subdomain

    @property
    def _gvol(self):
        return super(Curve, self).gvol

    @property
    def thickness(self):
        if self.axis is not None:
            return 2*np.pi * self.radius
        if not hasattr(self, '_thickness'):
            self._thickness = gScalar(self, 1.)
        return self._thickness

    @property
    def width(self):
        if not hasattr(self, '_width'):
            self._width = gScalar(self, 1.)
        return self._width

    @builtins.property
    def cross(self):
        return self.thickness * self.width

    @builtins.property
    def gvol(self):
        return self._gvol * self.cross


class Points(BaseDomain):
    """
    0D points (dummy elements)
    """
    def __getitem__(self, key):
        subdomain = BaseDomain.__getitem__(self, key)
        subdomain.gvol[:] = self.gvol[key]
        subdomain.axis = self.axis
        return subdomain

    @property
    def gvol(self):
        if self.axis is not None:
            return 2*np.pi * self.radius
        if not hasattr(self, '_gvol'):
            self._gvol = gScalar(self, 1.)
        return self._gvol


class Surface3D(Surface):
    """
    3D surface with tangents and normals
    """
    @property
    def faces(self):
        faces = super(Surface3D, self).faces
        if faces is not None:
            t = np.einsum('eni, bgnj -> ebgij',
                          self.ecoors, self.etype.bgshpd)
            n = normalize(np.cross(t[..., 0], t[..., 1]))
            n = n.reshape(-1, *n.shape[2:])
            # Agree rows, might be necessary if domain is a subdomain
            #
            # TODO This is a mess which should not be here. Currently this
            # agreement is necessary only here so maybe this is not big deal
            # however this might arise also in other places too. The general
            # problem is that domain faces elements can be given in any
            # order when subdomain is created (faces are also subdomains),
            # but here we need a specific order related to domain conec.
            # It seems there's no obvious way to assure globally that
            # domain.faces.eidx.conec_g is always coincident with
            # domain.eidx.conec_g[:, self.etype.faces]
            # Maybe didx, fidx should be computed when subdomain is created?
            fconecd = self.eidx.conec_g[:, self.etype.faces]
            fconecd = fconecd.reshape(-1, fconecd.shape[-1])
            uidx = tools.unique_rows(fconecd, strict=False)
            fconecd = fconecd[uidx]
            n = n[uidx]
            fconecf = faces.eidx.conec_g
            # ## is this really necesary ??
            didx, fidx = tools.agree_rows(fconecd, fconecf)
            # ##
            # Assign orientarion to faces
            # faces.orientation[:] = n.reshape(faces.orientation.shape)
            faces.orientation[fidx] = n.reshape(faces.orientation.shape)[didx]
        return faces

    @property
    def gjac(self):
        t = np.einsum('eni, gnj -> egij', self.ecoors, self.etype.gshpd)
        n = normalize(np.cross(t[..., 0], t[..., 1]))
        gjac = gTensor(self)
        gjac[..., :2] = t
        gjac[..., 2] = n
        return gjac

    @property
    def ttn(self):
        uvw = self.uvw
        t1 = gVector(self, uvw[..., 0])
        t2 = gVector(self, uvw[..., 1])
        n = gVector(self, uvw[..., 2])
        return t1, t2, n

    @property
    def tn(self):
        uvw = self.uvw
        t = gMatrix(self, uvw[..., 0:2])
        n = gVector(self, uvw[..., 2])
        return t, n

    @property
    def tangent(self):
        return gMatrix(self, self.uvw[..., 0:2])

    @property
    def tangent1(self):
        return gVector(self, self.uvw[..., 0])

    @property
    def tangent2(self):
        return gVector(self, self.uvw[..., 1])

    @property
    def normal(self):
        return gVector(self, self.uvw[..., 2])


class Curve2D(Curve):
    """
    2D curve with tangent and normal
    """
    rot90 = np.array([[0., -1.], [1., 0.]])  # counterclockwise (inward normal)
    # rot90 = np.array([[0., 1.], [-1., 0.]])  # clockwise (outward normal)

    @property
    def gjac(self):
        t = np.einsum('eni, gnj -> egi', self.ecoors, self.etype.gshpd)
        n = normalize(np.einsum('egi, ki -> egk', t, self.rot90))
        gjac = gTensor(self)
        gjac[..., 0] = t  # if t will be normalized it will be a rotation m.
        gjac[..., 1] = n
        return gjac

    @property
    def tn(self):
        uvw = self.uvw
        t = gVector(self, uvw[..., 0])
        n = gVector(self, uvw[..., 1])
        return t, -n  # "-" for outward normal

    @property
    def tangent(self):
        return gVector(self, self.uvw[..., 0])

    @property
    def normal(self):
        return -gVector(self, self.uvw[..., 1])  # "-" for outward normal


class Curve3D(Curve):
    """
    3D curve with orientation tangent and normal
    """
    def __getitem__(self, key):
        subdomain = Curve.__getitem__(self, key)
        subdomain.orientation[:] = self.orientation[key]
        # TODO: make a special function to get data only # .getdata(key)
        return subdomain

    @property
    def orientation(self):
        if not hasattr(self, '_orientation'):
            self._orientation = gVector(self, np.random.rand(3))
        return self._orientation
    # orientation = np.random.rand(3)  # OK for circular cross sections

    @property
    def gjac(self):
        t = np.einsum('eni, gnj -> egi', self.ecoors, self.etype.gshpd)
        # Deal with orientation a bit
        o = np.asarray(self.orientation, float)
        if o.shape == (self.nelem, self.etype.ngauss, 3):
            pass
        elif o.shape == (self.etype.ngauss, 3):
            o = np.tile(o, (self.nelem, 1, 1))
        elif o.shape == (3,):
            o = np.tile(o, (self.nelem, self.etype.ngauss, 1))
        else:
            raise ValueError("Wrong 'orientation' shape: %s" % o.shape)
        n = normalize(np.cross(t, o))
        o = normalize(np.cross(t, -n))
        gjac = gTensor(self)
        gjac[..., 0] = t
        gjac[..., 1] = o
        gjac[..., 2] = n
        return gjac

    @property
    def ton(self):
        uvw = self.uvw
        t = gVector(self, uvw[..., 0])
        o = gVector(self, uvw[..., 1])
        n = gVector(self, uvw[..., 2])
        return t, o, n

    @property
    def tangent(self):
        return gVector(self, self.uvw[..., 0])

    @property
    def orient(self):
        return gVector(self, self.uvw[..., 1])

    @property
    def normal(self):
        return gVector(self, self.uvw[..., 2])


# Set faces types
Volume.btype = Surface3D
Surface3D.btype = Curve3D
Curve3D.btype = Points
Surface.btype = Curve2D
Curve2D.btype = Points
Curve.btype = Points
Points.btype = None


# Domain type guessing
def guess_dtype(ndime, edime):
    # Choose proper domain class and class for boundary
    if ndime == 3:
        if edime == 3:
            cls = Volume
        elif edime == 2:
            cls = Surface3D
        elif edime == 1:
            cls = Curve3D
        elif edime == 0:
            cls = Points
        else:
            raise ValueError("Cannot create %iD domain with %iD elements"
                             % (ndime, edime))
    elif ndime == 2:
        if edime == 2:
            cls = Surface
        elif edime == 1:
            cls = Curve2D
        elif edime == 0:
            cls = Points
        else:
            raise ValueError("Cannot create %iD domain with %iD elements"
                             % (ndime, edime))
    elif ndime == 1:
        if edime == 1:
            cls = Curve
        elif edime == 0:
            cls = Points
        else:
            raise ValueError("Cannot create %iD domain with %iD elements"
                             % (ndime, edime))
    else:
        raise ValueError("Sorry, %iD domains are not supported" % ndime)
    return cls


if __name__ == '__main__':
    from fempy.testing import testmodule
    testmodule(__file__, '-v')



# -*- coding: utf-8 -*-
"""
Created on Sat Oct  5 10:31:10 2013

@author: mwojc
"""

import numpy as np
import networkx as nx
from fempy.geometry.gmsh import Gmsh
from fempy.geometry import tools
import copy


class Geometry(object):
    """
    Class representing 3D geometry.

    Attributes
    ----------
    points : list
        Points representing geometry
    curves : list
        Curves representing geometry
    surfaces: list
        Surfaces representing geometry
    volumes: list
        Volumes representing geometry
    coors: numpy array
        Coordinates of points
    gmsh : fempy.geometry.Gmsh instance
        Generator of finite element meshes
    elsize : float or None
        Parameter determinig element size at all points of geometry
    elsize_factor: float or None
        Number by which elsize at each point is multiplied

    Methods
    -------

    """
    tol = 1e-7

    def __init__(self, entities=[], **kwargs):
        self._entities = set(entities)
        if 'elsize' not in kwargs:
            self.__dict__['elsize'] = None
        if 'elsize_factor' not in kwargs:
            self.__dict__['elsize_factor'] = None
        for (k, v) in kwargs.items():
            setattr(self, k, v)

    def __repr__(self):
        return '%s' % (self.__class__.__name__,)  # self.groups)

    def __add__(self, other):
        return self.add(other)

    def __sub__(self, other):
        return self.remove(other)

    def __setattr__(self, name, value):
        if name == 'elsize':
            # set via __dict__ to avoid recurrency
            for e in self.points + self.curves + self.surfaces + self.volumes:
                e.__dict__['elsize'] = value
                if not hasattr(e, 'elsize_factor'):
                    e.__dict__['elsize_factor'] = 1.0
                if e.elsize_factor is None:
                    e.__dict__['elsize_factor'] = 1.0
        elif name == 'elsize_factor':
            for e in self.points + self.curves + self.surfaces + self.volumes:
                e.__dict__['elsize_factor'] = value
                if e.elsize is None:
                    e.__dict__['elsize'] = 1.0
        else:
            object.__setattr__(self, name, value)

    def __get_groups(self, entities):
        groups = {}
        for ent in entities:
            for group in ent.groups:
                if group not in groups:
                    groups[group] = [ent]
                else:
                    groups[group] += [ent]
        return groups

    @property
    def points(self):
        pts = []
        for entity in self._entities:
            if isinstance(entity, Point):  # infinite reccurency otherwise
                pts += [entity]
            else:
                pts += entity.points
        #return list(set(pts))   # TODO: order is lost!
        return tools.uniquify(pts)

    @property
    def curves(self):
        crv = []
        for entity in self._entities:
            if isinstance(entity, Point):
                pass
            elif isinstance(entity, Curve):
                crv += [entity]
            else:
                crv += entity.curves
        #return list(set(crv))
        return tools.uniquify(crv)

    @property
    def surfaces(self):
        srf = []
        for entity in self._entities:
            if isinstance(entity, (Point, Curve)):
                pass
            elif isinstance(entity, Surface):
                srf += [entity]
            else:
                srf += entity.surfaces
        #return list(set(srf))
        return tools.uniquify(srf)

    @property
    def volumes(self):
        return []

    @property
    def point_groups(self):
        return self.__get_groups(self.points)

    @property
    def curve_groups(self):
        return self.__get_groups(self.curves)

    @property
    def surface_groups(self):
        return self.__get_groups(self.surfaces)

    @property
    def volume_groups(self):
        return self.__get_groups(self.volumes)

    @property
    def coors(self):
        coors = []
        for p in self.points:
            coors.append(p.coors)
        return np.array(coors)

    @property
    def gmsh(self):
        if not hasattr(self, '__gmsh'):
            self.__gmsh = Gmsh(self)
        return self.__gmsh

    def eq(self, other):
        return self == other

    def ne(self, other):
        return not self.eq(other)

    def add(self, others):
        """
        This method always returns new Geometry object
        """
        if not isinstance(others, (list, tuple)):
            others = [others]
        entities = self._entities
        sname = self.__class__.__name__
        for other in others:
            if not isinstance(other, Geometry):
                oname = other.__class__.__name__
                raise TypeError('Cannot add %s to %s' % (oname, sname))
            entities = entities.union(other._entities)
        return Geometry(entities)

    def remove(self, others):
        """
        This method always returns new Geometry object
        """
        if not isinstance(others, (list, tuple)):
            others = [others]
        entities = self._entities
        sname = self.__class__.__name__
        for other in others:
            if not isinstance(other, Geometry):
                oname = other.__class__.__name__
                raise TypeError('Cannot remove %s from %s' % (oname, sname))
            entities = entities.difference(other._entities)
        return Geometry(entities)

    def copy(self):
        return copy.deepcopy(self)

    def save(self, fname):
        tools.saveobj(self, fname, ext='.Geometry')

    @classmethod
    def load(cls, fname):
        return tools.loadobj(fname, ext='.Geometry')

    def move(self, vec):
        for p in self.points:
            p.move(vec)

    def rotate(self, angle, axis=[0., 0., 1.]):
        # angle in radians
        R = tools.rotation_matrix(axis, angle)
        for p in self.points:
            p._rotate(R)

    def resize(self, m=1., center=None, resize_elsize_factor=True):
        coors = self.coors
        if center is None:
            center = coors.mean(axis=0)
        vectors = coors - center
        updates = vectors * (m - 1.)
        for p, vec in zip(self.points, updates):
            p.move(vec)
        if resize_elsize_factor:
            if self.elsize_factor is not None:
                self.elsize_factor = self.elsize_factor * m


class Primitive(object):
    @property
    def groups(self):
        try:
            return sorted(list(self.__groups))
        except AttributeError:
            return []

    def add_to_groups(self, groups):
        if isinstance(groups, str):
            groups = groups.split(',')
            groups = [g.strip() for g in groups]
        groups = [str(g) for g in groups]
        try:
            self.__groups.update(groups)
        except:
            self.__groups = set(groups)

    def remove_from_groups(self, groups):
        if isinstance(groups, str):
            groups = groups.split(',')
            groups = [g.strip() for g in groups]
        self.__groups.difference_update(groups)

    def clear_groups(self):
        self.__groups.clear()


class Point(Geometry, Primitive):
    def __init__(self, coors, groups=[], **kwargs):
        if 'tol' in kwargs:
            self.tol = kwargs['tol']
        self.__coors = tools.coors3D(coors)
        self.add_to_groups(groups)
        Geometry.__init__(self, [self], **kwargs)

    @property
    def coors(self):
        return self.__coors

    def _rotate(self, R):
        self.__coors = R.dot(self.__coors)

    def move(self, vec):
        vec = tools.coors3D(vec)
        self.__coors += vec

    def eq(self, other):
        if type(self) == type(other):
            tol = min(self.tol, other.tol)
            return (abs(self.coors - other.coors) < tol).all()
        return False

    def distance(self, other):
        return np.linalg.norm(self.coors-other.coors)


class Curve(Geometry, Primitive):
    orientation = 1  # physical orientation - manipulated in PlaneSurface
    slaves = None

    def eq(self, other):
        if type(self) == type(other):
            sp = self.points
            op = other.points
            if len(sp) == len(op):  # not necesserily true for splines
                check = [p1.eq(p2) for p1, p2 in zip(sp, op)]
                if False not in check:
                    return True
        return False

    def is_reversed(self, other):
        if type(self) == type(other):
            sp = self.points
            op = other.points[::-1]
            check = [p1.eq(p2) for p1, p2 in zip(sp, op)]
            if False not in check:
                return True
        return False

    def periodic(self, curve, sign=-1):
        if self.slaves is None:
            self.slaves = {curve: sign}
        else:
            self.slaves[curve] = sign


class Line(Curve):
    def __init__(self, start, end, groups=[], **kwargs):
        if 'tol' in kwargs:
            self.tol = kwargs['tol']
        assert start.ne(end)
        self.__start = start
        self.__end = end
        self.add_to_groups(groups)
        Geometry.__init__(self, [self], **kwargs)

    @property
    def start(self):
        return self.__start

    @property
    def end(self):
        return self.__end

    @property
    def points(self):
        return [self.start, self.end]

    @property
    def coors(self):
        return np.array([self.start.coors, self.end.coors])

    @property
    def length(self):
        return np.linalg.norm(self.start.coors-self.end.coors)

    @property
    def _lines(self):
        return [self]

    def poly(self, res=None):
        if res is None:
            res = 2
        return tools.multi_linspace(self.start.coors, self.end.coors, num=res)


class Spline(Curve):
    def __init__(self, points, groups=[], **kwargs):
        if 'tol' in kwargs:
            self.tol = kwargs['tol']
        if len(points) == 2:
            assert points[0].ne(points[1])
        self.__points = points  # List of points
        self.add_to_groups(groups)
        Geometry.__init__(self, [self], **kwargs)

    @property
    def start(self):
        return self.__points[0]

    @property
    def end(self):
        return self.__points[-1]

    @property
    def points(self):
        return self.__points

    @property
    def coors(self):
        coors = []
        for p in self.points:
            coors.append(p.coors)
        return np.array(coors)

    @property
    def is_closed(self):
        c = False
        if self.start.eq(self.end):
            c = True
        return c

    @property
    def _lines(self):
        p = self.points
        return [Line(p[i], p[i+1]) for i in range(len(p)-1)]

    def poly(self, res=None):
        if res is None:
            res = len(self.points)
        N = np.linspace(0, 1, res)
        x, y, z = self.coors.T
        ic = self.is_closed
        # interpolate
        x = tools.GmshSpline(x, closed=ic).eval(N)
        y = tools.GmshSpline(y, closed=ic).eval(N)
        z = tools.GmshSpline(z, closed=ic).eval(N)
        return np.array([x, y, z]).T

    # def length(self):  #AND SO ON...


class CircularArc(Curve):
    def __init__(self, start, center, end, groups=[], **kwargs):
        if 'tol' in kwargs:
            self.tol = kwargs['tol']
        assert start.ne(center)
        assert start.ne(end)
        assert center.ne(end)
        radius1 = center.distance(start)
        radius2 = center.distance(end)
        chord = start.distance(end)
        assert abs(radius1 - radius2) <= self.tol  # cocircular?
        assert radius1 - chord/2. >= self.tol  # half circle?
        self.__start = start
        self.__center = center
        self.__end = end
#        self.__midpoint = self.__get_midpoint()
        self.add_to_groups(groups)
        Geometry.__init__(self, [self], **kwargs)

    @property
    def start(self):
        return self.__start

    @property
    def center(self):
        return self.__center

    @property
    def end(self):
        return self.__end

    @property
    def midpoint(self):
        return self.__get_midpoint()

    @property
    def plane(self):
        return self.__get_plane()

    @property
    def points(self):
        return [self.start, self.center, self.end]

    @property
    def radius(self):
        return self.center.distance(self.start)

    @property
    def chord(self):
        return self.start.distance(self.end)

    @property
    def angle(self):
        return 2 * np.arcsin((self.chord/2.)/self.radius)

    @property
    def length(self):
        return self.angle*self.radius

    @property
    def _lines(self):
        midpoint = self.midpoint
        return [Line(self.start, midpoint), Line(midpoint, self.end)]

    def poly(self, res=None):
        if res is None:
            return np.array([self.start.coors, self.midpoint.coors,
                             self.end.coors])
        c = self.center.coors
        r = self.radius
        a = self.angle
        u, v = self.plane[:-1]
        ai = np.linspace(0, a, res)
        pcoors = r*np.einsum('n, i -> ni', np.cos(ai), u) +\
            r*np.einsum('n, i -> ni', np.sin(ai), v) + c
        return pcoors

    def __get_plane(self):
        c = self.center.coors
        sc = self.start.coors - c
        ec = self.end.coors - c
        u = sc/np.linalg.norm(sc)
        n = np.cross(sc, ec)
        n = n/np.linalg.norm(n)
        v = np.cross(n, u)
        return u, v, n

    def __get_midpoint(self):
        c = self.center.coors
        r = self.radius
        a = self.angle
        u, v = self.plane[:-1]
        mcoors = r*np.cos(a/2.)*u + r*np.sin(a/2.)*v + c
        return Point(mcoors)


class EllipticalArc(Curve):
    def __init__(self, start, center, major, end, groups=[], **kwargs):
        if 'tol' in kwargs:
            self.tol = kwargs['tol']
        self.__start = start
        self.__center = center
        self.__major = major
        self.__end = end
        assert center.ne(major)
        assert tools.coplanar(self.coors, self.tol)
        assert not tools.colinear([start.coors, center.coors, end.coors],
                                  self.tol)
        self.radiuses  # raises ValueError is radiuses cannot be calcukated
        self.add_to_groups(groups)
        Geometry.__init__(self, [self], **kwargs)

    @property
    def start(self):
        return self.__start

    @property
    def center(self):
        return self.__center

    @property
    def major(self):
        return self.__major

    @property
    def end(self):
        return self.__end

    @property
    def midpoint(self):
        return self.__get_midpoint()

    @property
    def points(self):
        return [self.start, self.center, self.major, self.end]

    @property
    def start_radius(self):
        return self.center.distance(self.start)

    @property
    def end_radius(self):
        return self.center.distance(self.end)

    @property
    def chord(self):
        return self.start.distance(self.end)

    @property
    def plane(self):
        return self.__get_plane()

    @property
    def start_polar_angle(self):
        v1 = self.major.coors - self.center.coors  # self.__get_plane()[0]
        v2 = self.start.coors - self.center.coors
        return tools.signed_angle(v1, v2)

    @property
    def end_polar_angle(self):
#        v1 = self.major.coors - self.center.coors  # self.__get_plane()[0]
#        v2 = self.end.coors - self.center.coors
#        return tools.signed_angle(v1, v2)
        return self.start_polar_angle + self.polar_angle

    @property
    def polar_angle(self):
        v1 = self.start.coors - self.center.coors
        v2 = self.end.coors - self.center.coors
        return tools.signed_angle(v1, v2)

    @property
    def start_parametric_angle(self):
        return self.__get_parametric_angle(self.start_polar_angle)

    @property
    def end_parametric_angle(self):
        return self.__get_parametric_angle(self.end_polar_angle)

    @property
    def parametric_angle(self):
        return self.end_parametric_angle - self.start_parametric_angle

    @property
    def eccentricity(self):
        A, B = self.radiuses
        a = max(A, B)
        b = min(A, B)
        return np.sqrt(a**2 - b**2)/a

    @property
    def length(self):
        a = max(self.radiuses)
        t1 = self.start_parametric_angle
        t2 = self.end_parametric_angle
        m = self.eccentricity**2
        from scipy.special import ellipeinc  # second kind
        l1 = a*ellipeinc(t1, m)
        l2 = a*ellipeinc(t2, m)
        return l2 - l1

    @property
    def radiuses(self):
        r_1 = self.start_radius
        r_2 = self.end_radius
        t_1 = self.start_polar_angle
        t_2 = self.end_polar_angle
        sqrt = np.sqrt
        sin = np.sin
        cos = np.cos
        # M = np.array([[np.cos(sa), np.sin(sa)], [np.cos(ea), np.sin(ea)]])
        # A, B = np.einsum('ij, j -> i', np.linalg.inv(M), [sr, er])
        usa = r_1**2*r_2**2*(-cos(2*t_1) + cos(2*t_2)) / \
            (r_1**2*sin(t_1)**2 - r_2**2*sin(t_2)**2)
        usb = r_1**2*r_2**2*(cos(2*t_1) - cos(2*t_2)) / \
            (r_1**2*cos(t_1)**2 - r_2**2*cos(t_2)**2)
        if usa < self.tol or usb < self.tol:
            raise ValueError(
                "Wrong ellipse definition. Radiuses cannot be calculated."
                )
        A = sqrt(2)*sqrt(usa)/2.
        B = sqrt(2)*sqrt(usb)/2.
        return A, B

    @property
    def _lines(self):
        midpoint = self.midpoint
        return [Line(self.start, midpoint), Line(midpoint, self.end)]

    def __get_plane(self):
        c = self.center.coors
        mc = self.major.coors - c
        ec = self.end.coors - c
        u = mc/np.linalg.norm(mc)
        n = np.cross(u, ec)
        if np.linalg.norm(n) == 0.:
            ec = self.start.coors - c
            n = np.cross(u, ec)
        if tools.signed_angle(u, ec) < 0.:  # make xy ccw
            n = -n
        n = n/np.linalg.norm(n)
        v = np.cross(n, u)
        return u, v, n  # u - major axis, v - minor axis

    def __get_parametric_angle(self, theta):
        A, B = self.radiuses
        n = int(theta // (np.pi/2.))
        if n % 2 == 0:
            frac = A/B  # we are in first or third quarter
        else:
            frac = B/A  # we are in second or fourth quarter
        base = n * np.pi/2.
        return base + np.arctan(frac * np.tan(theta-base))

    def __get_midpoint(self):
        c = self.center.coors
        A, B = self.radiuses
        # is this really midpoint ??
        a = (self.start_parametric_angle + self.end_parametric_angle)/2.
        u, v = self.plane[:-1]
        mcoors = A*np.cos(a)*u + B*np.sin(a)*v + c
        return Point(mcoors)

    def poly(self, res=None):
        if res is None:
            return np.array([self.start.coors, self.midpoint.coors,
                             self.end.coors])
        c = self.center.coors
        A, B = self.radiuses
        t1 = self.start_parametric_angle
        t2 = self.end_parametric_angle
        u, v = self.plane[:-1]
        ai = np.linspace(t1, t2, res)
        pcoors = A*np.einsum('n, i -> ni', np.cos(ai), u) +\
            B*np.einsum('n, i -> ni', np.sin(ai), v) + c
        return pcoors

    def eq(self, other):
        if type(self) == type(other):
            sp = self.start, self.midpoint, self.end
            op = other.start, other.midpoint, other.end
            if False not in [p1.eq(p2) for p1, p2 in zip(sp, op)]:
                return True
        return False

    def is_reversed(self, other):
        if type(self) == type(other):
            sp = self.start, self.midpoint, self.end
            op = other.end, other.midpoint, other.start
            if False not in [p1.eq(p2) for p1, p2 in zip(sp, op)]:
                return True
        return False


class Surface(Geometry, Primitive):
    pass


class PlaneSurface(Surface):
    def __init__(self, curves, groups=[], holes=[], orientation=1, **kwargs):
        if 'tol' in kwargs:
            self.tol = kwargs['tol']
        self.__graph, self.__curves, self.__corient = self.__arrange(curves)
        self.__holes = []
        self.add_holes(holes)
        if not tools.coplanar(self.coors, tol=self.tol):
            raise ValueError("Coordinates of points are not coplanar")
        self.add_to_groups(groups)
        self.orientation = orientation  # see __setattr__
        Geometry.__init__(self, [self], **kwargs)
        # TODO: what about covering lines?

    @property
    def curves(self):
        curves = [c for c in self.__curves]  # new list
        for h in self.holes:
            curves += h.curves  # recursive
        return curves

    @property
    def corient(self):
        corient = [o for o in self.__corient]  # new list
        for h in self.holes:
            corient += h.corient  # recursive, negative orients not work
        return corient

    @property
    def loops(self):
        loops = [(self.__curves, self.__corient)]
        loops += sum([h.loops for h in self.holes], [])  # recursive
        return loops

    @property
    def graph(self):
        return self.__graph

    @property
    def holes(self):
        return self.__holes

    @property
    def points(self):  # in current orientation order, no repetitions
        points = []
        for c, o in zip(self.curves, self.corient):
            if o == 1:
                pts = c.points
            else:
                pts = c.points[::-1]
            for p in pts:
                if p not in points:
                    points += [p]
        return points

#    @property
#    def coors(self):
#        coors = []
#        for p in self.points:
#            coors.append(p.coors)
#        return np.array(coors)
#
#    @property
#    def center(self):
#        return self.coors.mean(axis=0)

    def __arrange(self, curves):
        # Create graph
        g = nx.Graph()
        uvd = []
        for c in curves:
            for p in c._lines:
                uvd += [(p.start, p.end, {'curve': c})]
        g.add_edges_from(uvd)
        # Recreate graph with cycles only
        cycles = nx.cycle_basis(g)
        g = g.subgraph(sum(cycles, []))
        # Check if proper graph
        if len(cycles) == 0:
            raise ValueError('No line loops found')
        if len(cycles) > 1:
            degree = g.degree()
            if hasattr(degree, 'items'):  # in networkx-2.0 degree is not
                degree = degree.items()   # a dict any more ...
            for n, deg in degree:
                if deg > 2:
                    raise ValueError("To many lines in point: %s" % n)
        # Make some efforts to get predictible order of curves
        for u, v, d in uvd:
            if u in g:
                root = u
                after_root = v
                break
        cycle = nx.cycle_basis(g, root=root)[0][::-1]  # only one cycle exists
        #if cycle[-1] == after_root:
        if cycle[1] != after_root:
            cycle = [cycle[0]] + cycle[1:][::-1]  #
        # The above can be replaces by simple:
        #   cycle = nx.cycle_basis(g)[0][::-1]
        # but then we will get curves starting from random curve  (SO WHAT?)
        # Now get list of curves and orientations
        circuit = cycle + [cycle[0]]
        curves = []
        corient = []
        for i in range(len(cycle)):
            u = circuit[i]
            v = circuit[i+1]
            c = g[u][v]['curve']
            if c not in curves:  # This 'if' is for circle hack
                curves += [c]
                if c.start == u:  # This is always true for closed spline!
                    corient += [1]
                else:
                    corient += [-1]
        return g, curves, corient

    def __coplanar_curves(self):
        # coors computed twice here
        rank = np.linalg.matrix_rank(self.coors - self.center, tol=self.tol)
        if rank == 2:
            return True
        return False

    def __get_orientation(self):
        poly = self.poly()
        pcenter = poly[:-1].mean(axis=0) + 10*self.tol  # shift a bit to
                                                        # avoid perfect symm.
        vectors = poly - pcenter
        lengths = np.sqrt((vectors**2).sum(1))
        slengths = lengths[0:-1] * lengths[1:]
        dots = (vectors[0:-1] * vectors[1:]).sum(1)
        cross = np.cross(vectors[0:-1], vectors[1:])
        angles = np.arccos((dots/slengths).round(15))
        signs = np.sign(cross.sum(1))  # cross.sum(1): dot product with [1,1,1]
        if (angles*signs).sum() < 0:  # is possible = 0 here? no, if no symm.
            return -1  # clockwise
        return 1  # counter-clockwise

    def __setattr__(self, name, value):
        if name == 'orientation':
            if not hasattr(self, 'orientation'):
                current_orientation = self.__get_orientation()
            else:
                current_orientation = self.orientation

            if value not in [1, 'ccw', 'counterclockwise',
                             -1, 'cw', 'clockwise']:
                raise AttributeError('Wrong orientation name')

            curves = self.__curves
            corient = self.__corient
            c1 = value in [1, 'ccw', 'counterclockwise'] and \
                current_orientation in [-1, 'cw', 'clockwise']
            c2 = value in [-1, 'cw', 'clockwise'] and \
                current_orientation in [1, 'ccw', 'counterclockwise']
            if c1 or c2:
                self.__curves = [curves[0]] + curves[1:][::-1]
                self.__corient = [corient[0]] + corient[1:][::-1]
                self.__corient = [-o for o in self.__corient]
            for c, o in zip(self.__curves, self.__corient):
                c.orientation = o   # This can make also
                                    # unwanted changes!
        Geometry.__setattr__(self, name, value)

    def poly(self, res=None):  # repeated point at start and end
        pcoors = []
        #only exterior boundary
        for c, o in zip(self.__curves, self.__corient):
            poly = c.poly(res)
            if o == -1:
                poly = poly[::-1]
            pcoors += poly[:-1].tolist()
        if Point(pcoors[0]).ne(Point(pcoors[-1])):
            pcoors = np.asarray(pcoors + [pcoors[0]])  # close polygon
        return pcoors

    def poly_center(self, res=None):
        poly = self.poly(res=res)
        return poly.mean(0)

    def poly_bbox(self, res=None):
        poly = self.poly(res=res)
        mini = poly[poly.argmin(axis=0)]
        maxi = poly[poly.argmax(axis=0)]
        return mini, maxi

    def add_hole(self, hole, validate=False, m=1.0, res=50):
        ss = self
        os = hole
        add_hole = True
        if validate:
            ost = os.copy()
            ost.resize(m)
            if ss.contains(ost):
                for ssh in ss.holes:
                    if not ost.disjoint(ssh, res) or \
                       not ssh.disjoint(ost, res):
                        add_hole = False
                        break
            else:
                add_hole = False
        if add_hole:
            ss.__holes = ss.__holes + [os]
            return True
        return False

    def add_holes(self, holes, **kwargs):
        for i, h in enumerate(holes):
            success = self.add_hole(h, **kwargs)
            if not success:
                raise ValueError('Hole %i cannot be added' % i)

    def remove_holes(self, holes):
        for h in holes:
            idx = self.__holes.index(h)
            del self.__holes[idx]

    def contains(self, other, res=50):
        polygon = self.poly(res)
        points = other.poly(res)
        test = tools.contains(polygon, points)
        return test

    def disjoint(self, other, res=50):
        polygon = self.poly(res)
        points = other.poly(res)
        test = tools.disjoint(polygon, points)
        return test


# Tests
if __name__ == '__main__':
    from fempy.testing import testmodule
    testmodule(__file__, '-v')



# -*- coding: utf-8 -*-

if __name__ == '__main__':
    import pytest
    import sys
    import os

    testdir, _ = os.path.split(__file__)

    if len(sys.argv) > 1:
        testdir = os.path.join(testdir, sys.argv[1])
    pytest.main([testdir, '-v'])  # Run tests



# -*- coding: utf-8 -*-
"""
Created on Fri Jun 10 19:09:58 2016

@author: mwojc
"""
from __future__ import print_function
import numpy as np
from scipy.optimize import newton_krylov
from scipy.interpolate import interp1d
from fempy.solvers import linear
from fempy.solvers import ivp


class NonLinearSolver(linear.IterativeSolver):
    def callback(self, x, f=None):
        if hasattr(self.model, 'callback'):
            self.model.callback(x, f)
        self.update(x)

    def nsolve(self, **options):
        self.lsolve(maxiter=1)
        # self.update()

        options = self.getopts(options)
        F = self.residual
        x0 = self.x
        method = self.method
        M = self.M

        x = newton_krylov(F, x0, method=method.method, inner_M=M,
                          callback=self.callback, **options)
        self.x = x

    def solve(self, steps=None, **options):
        m = self.model
        if steps is None or steps == 1:
            m.time = m.time1
            return self.nsolve(**options)
        elif isinstance(steps, int):
            steps = np.linspace(m.time0, m.time1, steps)
        else:
            assert steps[0] >= m.time0
            assert steps[-1] <= m.time1
            assert np.all(np.diff(steps) >= 0)
        x = []
        for step in steps:
            print("Step: {}".format(step))
            m.time = step
            self.nsolve(**options)
            x.append(self.x)
        # Interpolate solution for dense output
        # TODO: should we always solve for intermediate time step
        x = np.array(x).T
        sol = interp1d(steps, x, kind='linear', fill_value='extrapolate')
        sol.t = sol.x
        sol.sol = sol
        m._sol = sol
        m.__class__.sol = ivp._set_solution_at_time
        m.sol()  # This sets sotution at last time step


class NewtonKrylovSolver(NonLinearSolver):
    pass


class SimpleIterationSolver(NonLinearSolver):
    def nsolve(self, xtol=1e-5, maxiter=10, verbose=False, raise_error=False,
               **options):
        options = self.getopts(options)
        opts = {}
        for k, v in options.items():
            if 'inner_' in k:
                opts[k[6:]] = v
        self.lsolve(**opts)
        x0 = self.x
        for i in range(maxiter):
            self.update()
            self.lsolve(**opts)
            x = self.x.copy()
            dxnorm = np.linalg.norm(np.abs(x - x0))
            if verbose:
                print('{}:   xtol: {}'.format(i, dxnorm))
            if np.allclose(x, x0, rtol=xtol):
                return
            x0 = x.copy()
        if raise_error:
            raise ValueError("No convergence in %s iterations" % maxiter)
        else:
            import warnings
            warnings.warn(("No convergence in %s iterations" % maxiter))


if __name__ == '__main__':
    from fempy.testing import testmodule
    testmodule(__file__, '-v')



# -*- coding: utf-8 -*-
import pytest
from fempy.domain.tools import *
import fempy.testing as testing
import numpy as np


class TestTools(object):
    def test_unique_rows_lex(self):
        a = np.arange(12).reshape(4, 3)
        b = np.arange(12)[::-1].reshape(4, 3)
        c = np.concatenate((a, b))
        idx, msk = unique_rows_lex(c, sort=True, strict=False)
        assert np.allclose(idx, [0, 7, 1, 6, 2, 5, 3, 4])
        assert np.allclose(msk, [1, 0, 1, 0, 1, 0, 1, 0])

    def test_unique_rows(self):
        a = np.arange(12).reshape(4, 3)
        b = np.arange(12)[::-1].reshape(4, 3)
        c = np.concatenate((a, b))
        c2 = np.concatenate((b, a))
        assert np.allclose(unique_rows(c), True)
        assert np.allclose(unique_rows(c, strict=False), True)
        assert np.allclose(unique_rows(c, sort=True), False)
        assert np.allclose(unique_rows(c, sort=True, strict=False),
                           [True]*4 + [False]*4)
        assert np.allclose(unique_rows(c2, sort=True, strict=False),
                           [True]*4 + [False]*4)
        assert len(unique_rows(np.array([], dtype=np.int))) == 0

    def test_duplicate_rows(self):
        a = np.arange(12).reshape(4, 3)
        b = np.arange(12)[::-1].reshape(4, 3)
        c = np.concatenate((a, b))
        assert np.allclose(np.sort(duplicate_rows(c)),
                           [1, 2, 3, 4, 5, 6, 7, 8])
        assert np.allclose(duplicate_rows(c, sort=True),
                           [1, 2, 3, 4, 4, 3, 2, 1])
        assert len(duplicate_rows(np.array([], dtype=np.int))) == 0

    def test_in2d(self):
        a = np.arange(12).reshape(4, 3)
        b = np.arange(12)[::-1].reshape(4, 3)
        c = np.concatenate((a, b))
        np.allclose(in2d(c, a[:2]), [True]*2 + [False]*6)
        np.allclose(in2d(c, a[:2], sort=True),
                    [True]*2 + [False]*4 + [True]*2)

    def test_sum_rows(self):
        # integer arr
        arr = np.array([1, 2, 3, 4, 5, 6, 7, 8])
        mask = np.array([True, False, False, True, True,
                         False, True, False])
        res = np.array([1+2+3, 4, 5+6, 7+8])
        assert np.allclose(res, uneven_sum(arr, mask))
        # with keepshape
        res = np.array([6, 6, 6, 4, 11, 11, 15, 15])
        assert np.allclose(res, uneven_sum(arr, mask, keepshape=True))

        # boolean arr
        arr = np.array([True, False, True, False, False, True, False, True])
        mask = np.array([True, False, False, True, True, False, True, False])
        res = np.array([True, False, True, True])
        assert np.allclose(res, uneven_sum(arr, mask))
        res2 = np.add.reduceat(arr, np.nonzero(mask)[0])
        assert np.allclose(res2 > 0, uneven_sum(arr, mask))
        # with keepshape
        res = np.array([True, True, True, False, True, True, True, True])
        assert np.allclose(res, uneven_sum(arr, mask, keepshape=True))

    def test_agree_rows(self):
        a = np.arange(12).reshape(4, 3)
        b = np.arange(12)[::-1].reshape(4, 3)
        aidx, bidx = agree_rows(a, a[::-1])
        assert np.allclose(a[aidx], a[::-1][bidx])
        with pytest.raises(ValueError):
            aidx, bidx = agree_rows(a, b)

    def test_create_faces(self):
        d = testing.common.SquareWithCircle()
        f = create_faces(d)
        f['BOUNDARY']


if __name__ == '__main__':
    pytest.main([str(__file__), '-v'])  # Run tests from current file



# -*- coding: utf-8 -*-
"""
Created on Tue Jun 28 11:17:22 2016

@author: mwojc
"""

if __name__ == '__main__':
    import pytest
    pytest.main([str(__file__), '-v'])  # Run tests from current file



# -*- coding: utf-8 -*-
"""
Created on Sat Oct  6 11:43:13 2018

@author: awitek, mwojc
"""
import numpy as np
from scipy import sparse
from fempy.models.model import Model
from fempy.fields.nfields import nScalar
from fempy.fields.gfields import gScalar, gTensor
from fempy.fields.operators import dot, grad, operator, onboundary, nabla
from fempy.fields.operators import qform, lform, integrate, ongauss, lformu
from fempy.solvers import ivp


class HeatAndMoisture(Model):
    """ Kunzel model """
    default_solver = ivp.BDFSolver

    def init(self, *args, **kwargs):
        d = self.domain
        b = self.domain.boundary
        # on domain nodes
        self.T = nScalar(d, 273.15)     # temperature [K]
        self.fi = nScalar(d, 0.5)       # relative humidity [-]
        self.bcsT = nScalar(d, False)   # Dirichlet boundary conditions mask
        self.bcsfi = nScalar(d, False)  # Dirichlet boundary conditions mask
        # on domain integration points
        # heat
        self.kij = gTensor(d)           # heat conductivity          [W/(m*K)]
        self.cp_s = gScalar(d, 1.)      # solid heat capac. (isob.)  [J/(kg*K)]
        self.rho_s = gScalar(d, 1.)     # material density           [kg/m**3]
        self.h = gScalar(d)             # heat source                [W/kg]
        # moisture
        self.Dfi = gTensor(d)           # liquid conduction coeff.   [kg/(m*s)]
        self.dp = gTensor(d)            # water vapour permeab.   [kg/(m*s*Pa)]
        self.w = gScalar(d)             # water vapour resistance factor [-]
        self.ksi = gScalar(d)           # moisture storage capacity  [kg/m**3]
        self.rho_w = gScalar(d, 1000.)  # water density              [kg/m**3]
        self.cp_w = gScalar(d, 4190.)   # water heat capac. (isob.)  [J/(kg*K)]
        # on boundary
        self.qT = gScalar(b)            # RHS explicit thermal flux  [W/m**2]
        self.qfi = gScalar(b)           # RHS exp. moisture flux  [kg/(m*s**2)]
        self.Text = gScalar(b, 273.15)  # external temperature on boundary [K]
        self.fiext = gScalar(b, 0.5)    # external r. humidity on boundary [-]
        self.alpha = gScalar(b)         # heat transfer coef.      [W/(m**2*K)]
        self.beta = gScalar(b)          # moisture transfer coef. [kg/(m**2*K)]
        # constants
        self.hv = 2257.e3               # evaporation enthalpy of water [J/kg]
        # self.sigma = 5.67036713131e-8   # Stefan-Boltzmann  [W/(m**2 * K**4)]
        # self.Mw = 0.01801528            # molar weight of water      [kg/mol]
        # self.R = 8.3144598              # universal gas constant  [J/(mol*K)]
        # self.PL = 101325.               # atmosferic pressure        [Pa]
        # solution related attributes
        self.mult = 1.0                 # multiplicator of humidity equation

    @property
    def lhs(self):
        return np.concatenate((self.T.ravel(), self.fi.ravel()))

    @lhs.setter
    def lhs(self, lhs):
        n = self.T.nsize
        shape = self.T.shape
        self.T[:] = lhs[:n].reshape(shape)
        self.fi[:] = lhs[n:].reshape(shape)

    @property
    def bcs(self):
        return np.concatenate((self.bcsT.ravel(), self.bcsfi.ravel()))

    @property
    def rhsT(self):
        T = self.T
        Tb = onboundary(T)
        # Assemble sources
        source = dot(self.rho_s, self.h)
        qs = lform(source, operator(T))
        # Assemble convective heat transfer on boundary
        transfer = dot(self.alpha, self.Text)
        qt = lform(transfer, operator(Tb))
        return lform(self.qT) + qs + qt

    @property
    def rhsfi(self):
        fi = self.fi
        fib = onboundary(fi)
        # Assemble convective moisture transfer on boundary
        transfer = dot(self.betaext, self.fiext)
        qt = lform(transfer, operator(fib))
        return (lform(self.qfi) + qt) * self.mult

    @property
    def rhs(self):
        return np.concatenate((self.rhsT.ravel(), self.rhsfi.ravel()))

    def _psat(self, T):
        return np.exp(77.3450 + 0.0057*T - 7235./T)/T**8.2

    @property
    def psat(self):  # TODO: for T<0
        T = ongauss(self.T)
        return self._psat(T)

    @property
    def psatsurf(self):
        T = ongauss(onboundary(self.T))
        return self._psat(T)

    @property
    def psatext(self):
        T = self.Text
        return self._psat(T)

    @property
    def betasurf(self):
        return self.beta*self.psatsurf

    @property
    def betaext(self):
        return self.beta*self.psatext

    @property
    def dpsatdT(self):
        return 7235./(ongauss(self.T))**2 + 0.0057

    @property
    def k11(self):
        c = self.hv * ongauss(self.fi) * self.dpsatdT
        return self.kij + dot(self.dp, c)

    @property
    def k12(self):
        c = self.hv * self.psat
        return dot(self.dp, c)

    @property
    def k21(self):
        c = ongauss(self.fi) * self.dpsatdT
        return dot(self.dp, c)

    @property
    def k22(self):
        c = self.psat
        return self.Dfi + dot(self.dp, c)

    @property
    def jac(self):
        K11 = qform(self.k11, nabla(self.T))
        K11 += qform(self.alpha, operator(onboundary(self.T)))
        K22 = qform(self.k22, nabla(self.fi))
        K22 += qform(self.betasurf, operator(onboundary(self.fi)))
        K12 = qform(self.k12, nabla(self.fi))
        K21 = qform(self.k21, nabla(self.T))
        K22 *= self.mult
        K21 *= self.mult
        return sparse.bmat([[K11, K12], [K21, K22]])

    @property
    def gradT(self):
        return grad(self.T)

    @property
    def gradfi(self):
        return grad(self.fi)

    @property
    def QT(self):
        return -dot(self.k11, self.gradT)

    @property
    def Qfi(self):
        return -dot(self.k22, self.gradfi)

    @property
    def QTfi(self):
        return -dot(self.k12, self.gradfi)

    @property
    def QfiT(self):
        return -dot(self.k21, self.gradT)

    @property
    def lopT(self):
        LT1 = -integrate(self.QT)
        LT2 = lform(dot(self.alpha, ongauss(onboundary(self.T))))
        LT3 = -integrate(self.QTfi)
        return (LT1 + LT2 + LT3).ravel()

    @property
    def lopfi(self):
        Lfi1 = -integrate(self.Qfi)
        Lfi2 = lform(dot(self.betasurf, ongauss(onboundary(self.fi))))
        Lfi3 = -integrate(self.QfiT)
        return (Lfi1 + Lfi2 + Lfi3).ravel() * self.mult

    @property
    def lop(self):
        return np.concatenate((self.lopT, self.lopfi))

    @property
    def jactT(self):
        return lformu(self.rho_s * self.cp_s + self.w * self.cp_w)

    @property
    def jactfi(self):
        return lformu(self.ksi) * self.mult

    @property
    def jact(self):
        return np.concatenate((self.jactT.ravel(), self.jactfi.ravel()))


if __name__ == '__main__':
    # from fempy.geometry.shapes import Rectangle
    # d = Rectangle().gmsh.domain()
    # m = HeatAndMoisture(d)
    from fempy.testing import testmodule
    testmodule(__file__, '-v')



# -*- coding: utf-8 -*-
"""
Created on Wed Nov 19 10:49:39 2014

@author: mwojc
"""
import numpy as np
from fempy.fields.fields import Field
from fempy.fields.algebra import transpose, tensordot, dot, TensorCalc
import pytest


class Test_Algebra:
    s = np.random.rand(1)[0]
    v = np.random.rand(3)
    t = np.random.rand(3, 3)
    t3 = np.random.rand(3, 3, 3)
    t4 = np.random.rand(3, 3, 3, 3)
    a = np.random.rand(3, 2)

    fs = Field(np.random.rand(20, 10), 2)
    fv = Field(np.random.rand(20, 10, 3), 2)
    ft = Field(np.random.rand(20, 10, 3, 3), 2)
    ft3 = Field(np.random.rand(20, 10, 3, 3, 3), 2)
    ft4 = Field(np.random.rand(20, 10, 3, 3, 3, 3), 2)
    fa = Field(np.random.rand(20, 2, 3), 1)

    def test_transpose(self):
        tr = transpose
        nptr = np.transpose
        assert np.allclose(tr(self.t4), nptr(self.t4))
        assert np.allclose(tr(self.t3, (0, 2, 1)), nptr(self.t3, (0, 2, 1)))
        assert np.allclose(tr(self.ft4)[0, 0], nptr(self.ft4[0, 0]))
        assert np.allclose(tr(self.fa)[5], nptr(self.fa[5]))

    def test_tensordot(self):
        f = tensordot
        npf = np.tensordot
        assert np.allclose(f(self.t4, self.t), npf(self.t4, self.t))
        assert np.allclose(f(self.ft4, self.ft)[0, 0],
                           npf(self.ft4[0, 0], self.ft[0, 0]))
        assert np.allclose(f(self.fa, self.a, [-1, 0])[0],
                           npf(self.fa[0], self.a, [-1, 0]))
        assert np.allclose(f(self.fa, self.a, [[-1, -2], [0, 1]])[0],
                           npf(self.fa[0], self.a, [[-1, -2], [0, 1]]))
        assert np.allclose(f(self.fa, transpose(self.fa), [-1, -2])[0],
                           np.dot(self.fa[0], transpose(self.fa[0])))

        assert type(f(self.ft4, self.ft)) == Field
        assert type(f(self.t4, self.t)) == np.ndarray

        with pytest.raises(IndexError):
            f(self.fs, self.fv)
        assert np.allclose(f(self.fs, self.fv, 0), f(self.fv, self.fs, 0))
        assert np.allclose(f(self.fv, self.v, 1), f(self.v, self.fv, 1))

    def test_dot(self):
        f = dot
        npf = np.dot
        assert np.allclose(f(self.t4, self.t), npf(self.t4, self.t))
        assert np.allclose(f(self.ft4, self.ft)[0, 0],
                           npf(self.ft4[0, 0], self.ft[0, 0]))
        assert np.allclose(f(self.fa, self.a)[0],
                           npf(self.fa[0], self.a))
        assert np.allclose(f(self.ft, self.fv)[0, 0],
                           npf(self.ft[0, 0], self.fv[0, 0]))
        assert np.allclose(f(self.ft, self.v)[0, 0],
                           npf(self.ft[0, 0], self.v))
        assert np.allclose(f(self.v, self.ft)[0, 0],
                           npf(self.v, self.ft[0, 0]))
        assert np.allclose(f(self.v, self.fv)[0, 0],
                           npf(self.v, self.fv[0, 0]))

        assert np.allclose(f(self.fs, self.fv), f(self.fv, self.fs))
        assert np.allclose(f(self.fv, self.v), f(self.v, self.fv))
        assert np.allclose(f(self.fs, self.v)[0, 0],
                           npf(self.v, self.fs[0, 0]))


class TestTensorCalc:
    t3 = Field(np.random.rand(20, 10, 3, 3), 2)
    t2 = Field(np.random.rand(20, 10, 2, 2), 2)

    def call_calc(self, calc):
        calc.tensor
        calc.shape
        calc.T
        calc.sym
        calc.skew
        calc.diag
        calc.trace
        calc.mean
        calc.dev
        calc.pvals
        calc.pvecs
        calc.ptensor
        calc.polar
        calc.svd
        calc.eig
        calc.inv
        calc.det
        calc.I1
        calc.I2
        calc.I3
        calc.J2
        calc.J3
        calc.sqrtJ2
        calc.p
        calc.q
        calc.a
        calc.b
        calc.theta

    def test_call_t2(self):
        calc = TensorCalc(self.t2)
        self.call_calc(calc)

    def test_call_t3(self):
        calc = TensorCalc(self.t3)
        self.call_calc(calc)

    def test_call_t2sym(self):
        t2 = self.t2 + transpose(self.t2)
        calc = TensorCalc(t2)
        self.call_calc(calc)

    def test_call_t3sym(self):
        t3 = self.t3 + transpose(self.t3)
        calc = TensorCalc(t3)
        self.call_calc(calc)


if __name__ == '__main__':
    pytest.main([str(__file__), '-v'])  # Run tests from current file



# -*- coding: utf-8 -*-
"""
Created on Fri Feb 10 11:07:43 2017

@author: mwojc
"""
import numpy as np
import collections
import ast


def is_valid_variable_name(name):
    if '.' in name:  # or name.startswith('__'):  # or name in banned_keys:
        return False
    try:
        ast.parse('{} = None'.format(name))
        return True
    except:
        return False


def get_names_from_expression(expr):
    symbols = ['+', '-', '*', '/', '^',
               '**',
               '~',
               '==', '!=', '<', '<=', '>', '>=',
               '(', ')']
    for s in symbols:
        expr = expr.replace(s, ' ')
    evars = []
    for ev in expr.split():
        ev = ev.strip()
        if is_valid_variable_name(ev):
            evars.append(ev)
    return evars


def masked_index(n=None, lst=None):
    assert n is not None or lst is not None, "Please, give me 'n' or 'lst'"
    # Compute nlist, npos, nmask
    if lst is None:
        lst = np.arange(n)
    else:
        if isinstance(lst, np.ndarray):
            lst = [lst]
        lst = [np.asarray(l).ravel() for l in lst]
        lst = np.unique(np.concatenate(lst))
        if n is None:
            n = max(lst) + 1  # len(lst)
    pos = np.zeros(n, dtype=np.int) - 1
    pos[lst] = np.arange(lst.shape[0])
    mask = pos >= 0
    return lst, pos, mask


class Index(collections.MutableMapping):
    def _bool(self, value):
        mask = np.zeros(self.n, dtype=bool)
        if not isinstance(value, slice):
            value = np.asarray(value)
            if value.dtype != np.bool and value.dtype != np.int:
                value = value.astype(np.int)
        mask[value] = True
        return mask

    def _expr(self, key):
        # key = key.replace('-', '^')  # from numpy 1.14 '-' does not work
        names = get_names_from_expression(key)
        vdict = {'np': np}
        for n in names:
            v = self._get_value(n)
            if v is not None:
                vdict[n] = v
        try:
            value = eval(key, vdict)
        except Exception as e:
            raise KeyError("Processing key '%s' returned the following error: "
                           "%s: %s" % (key, type(e).__name__, e.args[0]))
        if not isinstance(value, np.ndarray) or \
            not value.dtype == np.bool or \
                not value.shape == (self.n,):
            raise KeyError("Key expression does not evaluate "
                           "to proper boolean mask: '%s, %s'" % (key, value))
        return value

    def _get_value_from_masks(self, key):
        if key in self._masks:
            return self._masks[key] * self.msk  # else None

    def _get_value_from_vars(self, key):
        if key in self._vars:
            return self._vars[key]  # else None

    def _get_value(self, key):
        # key is assumed to be a variable name
        value = self._get_value_from_masks(key)
        if value is None:
            value = self._get_value_from_vars(key)
        return value

    def __getitem__(self, key):
        if not isinstance(key, str):
            value = self._bool(key)
        else:
            value = self._get_value_from_masks(key)
            if value is None:
                value = self._expr(key)
        if len(value) == len(self.msk):  # TODO: only in global index
            value = value * self.msk
        return value

    def update(self, *E, **F):
        if len(E) and E[0] is None:
            return
        else:
            super(Index, self).update(*E, **F)


class GlobalNIndex(Index):
    def __init__(self, coors, lst=None):
        self.coors = coors
        self.lst, self.pos, self.msk = masked_index(len(coors), lst)
        self._masks = {}
        self._vars = {}
        for i in range(coors.shape[1]):
            self._vars['xyz'[i]] = coors[:, i]
        self.ng = len(coors)
        self.na = len(self.lst)
        self.n = self.ng

    def __setitem__(self, key, value):
        if not is_valid_variable_name(key):
            raise KeyError("Key not allowed: '%s'" % key)
        self._masks[key] = self._bool(value)

    # __getitem__ is defined in super class

    def __delitem__(self, key):
        del self._masks[key]

    def __iter__(self):
        return iter(self._masks)

    def __len__(self):
        return len(self._masks)


class NIndex(Index):
    def __init__(self, g, lst=None):  # lst is global!
        """
        Local node index
        """
        if isinstance(g, NIndex):
            raise ValueError("Sorry, you must use 'subindex' method "
                             "for creating NIndex from another one.")
        if isinstance(g, np.ndarray):
            g = GlobalNIndex(g, lst)
        self.g = g
        if lst is None:
            lst = self.g.lst
        lst = np.nonzero(self.g[lst])[0]  # ensures global msk is applied
        self.lst, self.pos, self.msk = masked_index(self.g.ng, lst)
        self.n = len(self.lst)
        # self.n = self.nd
        self._masks = {}
        self._vars = {}

    def __setitem__(self, key, value):
        self.g[key] = self.lst[value]

    def __delitem__(self, key):
        del self.g[key]

    def __iter__(self):
        return iter(self.g._masks)

    def __len__(self):
        return len(self.g._masks)

    def _get_value_from_masks(self, key):
        if key in self._masks:
            return self._masks[key]
        else:
            if key in self.g._masks:
                return self.g[key][self.msk]
        # return self.g._get_value_from_masks(key)[self.msk]

    def _get_value_from_vars(self, key):
        if key in self._vars:
            return self._vars[key]
        else:
            if key in self.g._vars:
                value = self.g._vars[key]
                if isinstance(value, np.ndarray):
                    value = value[self.msk]
                return value

    def set_local_mask(self, key, value):
        self._masks[key] = self._bool(value)

    def subindex(self, key=None, gkey=None):
        if key is None:  # if key is None returns a copy
            lst = self.lst
        else:
            lst = self.lst[self[key]]
        if gkey is not None:
            glst = np.nonzero(self.g[gkey])[0]
            lst = np.intersect1d(lst, glst, assume_unique=True)
        return self.__class__(self.g, lst)


class GlobalEIndex(GlobalNIndex):
    def __init__(self, conec, nidx, lst=None):
        self.conec = conec
        lst, pos, msk = masked_index(len(conec), lst)
        # but something can be shadowed by nidx
        if isinstance(nidx, NIndex):
            nidx = nidx.subindex(gkey=conec[lst])
        else:
            nidx = NIndex(nidx, conec[lst])
        lst = np.nonzero(nidx.msk[conec].all(1))[0]
        # ok, now we can assign
        self.lst, self.pos, self.msk = masked_index(len(conec), lst)
        self.nidx = nidx.subindex(gkey=conec[self.lst])
        self._masks = {}
        self._vars = {}
        self.ng = len(conec)
        self.na = len(lst)
        self.n = self.ng

    def __setitem__(self, key, value):
        super(GlobalEIndex, self).__setitem__(key, value)
        self.nidx.g[key] = self.conec[self._masks[key] * self.msk]

    def __getitem__(self, key):
        try:
            value = super(GlobalEIndex, self).__getitem__(key)
        except KeyError as e:
            try:  # this can potentially lead to unexpected results!
                value = self.nidx.g[key] * self.nidx.msk
                value = value[self.conec].all(1)
            except KeyError:
                raise e
        return value

    def __delitem__(self, key):
        del self._masks[key]
        try:
            del self.nidx[key]
        except KeyError:
            pass

    def __iter__(self):
        return iter(set(self._masks.keys() + self.nidx.keys()))

    def __len__(self):
        return len(set(self._masks.keys() + self.nidx.keys()))

    def _get_value_from_masks(self, key):
        if key in self._masks:
            return super(GlobalEIndex, self)._get_value_from_masks(key)
        else:
            if key in self.nidx.g._masks:
                return self.nidx.g[key][self.conec].all(1) * self.msk

    @property
    def nelem(self):
        return self.n

    @property
    def nenod(self):
        return self.conec.shape[1]


class EIndex(NIndex):
    """
    Local element index
    g       - GlobalEIndex instance or conec array
    lst     - list of elements from global index to be used
    nidx    - global or local node index, used only when 'g' is given as
              conec array (must be not None then)
    """
    def __init__(self, g, lst=None, nidx=None):
        # lst global!
        if isinstance(g, EIndex):
            raise ValueError("Sorry, use 'subindex' method "
                             "for creating EIndex from another one.")
        if isinstance(g, np.ndarray):
            if nidx is None:
                raise ValueError("'nidx' cannot be None if 'g' is a conec")
            g = GlobalEIndex(g, nidx, lst)
        self.g = g
        if lst is None:
            lst = self.g.lst
        lst = np.nonzero(self.g[lst])[0]  # ensures global msk is applied
        self.lst, self.pos, self.msk = masked_index(self.g.ng, lst)
        self.nidx = self.g.nidx.subindex(gkey=g.conec[self.lst])
        self.n = len(self.lst)
        # self.n = self.nd
        self._masks = {}
        self._vars = {}

    def __getitem__(self, key):
        try:
            value = super(EIndex, self).__getitem__(key)
        except KeyError as e:
            try:  # this can potentially lead to unexpected results!
                value = self.nidx.g[key] * self.nidx.msk
                value = value[self.g.conec[self.msk]].all(1)
            except KeyError:
                raise e
        return value

    def __delitem__(self, key):
        del self.g._masks[key]
        try:
            del self.nidx[key]
        except KeyError:
            pass

    def __iter__(self):
        return iter(set(self.g._masks.keys()).union(self.nidx.keys()))

    def __len__(self):
        return len(set(self.g._masks.keys()).union(self.nidx.keys()))

    def _get_value_from_masks(self, key):
        if (key in self.g._masks) or (key in self._masks):
            return super(EIndex, self)._get_value_from_masks(key)
        else:
            if (key in self.nidx.g._masks) or (key in self.nidx._masks):
                return self.nidx[key][self.conec].all(1)

    def set_local_mask(self, key, value):
        self._masks[key] = self._bool(value)
        # self.nidx[key] = self.conec[self._masks[key] * self.msk]
        self.nidx.set_local_mask(key, self.conec[self[key]])
        # .conec[self._masks[key] * self.msk])

    @property
    def conec_g(self):
        return self.g.conec[self.msk]

    @property
    def conec_a(self):
        return self.g.nidx.pos[self.conec_g]

    @property
    def conec(self):
        return self.nidx.pos[self.conec_g]

    @property
    def coors_g(self):
        return self.nidx.g.coors

    @property
    def coors_a(self):
        return self.nidx.g.coors[self.nidx.g.lst]

    @property
    def coors(self):
        return self.nidx.g.coors[self.nidx.lst]

    @property
    def nelem(self):
        return self.g.nelem

    @property
    def nenod(self):
        return self.g.nenod


class EGroups(collections.MutableMapping):
    def __init__(self, domain):
        self.domain = domain

    def __getitem__(self, key):
        if 'boundary' in key or 'BOUNDARY' in key or 'CUT' in key:
            self.domain.boundary  # creates boundary if possible and necessary
        group = self.domain.eidx[key]
        return group

    def __setitem__(self, key, value):
        d = self.domain
        d.eidx[key] = value
        f = d._faces
        if f is not None:  # and d.etype.ndime >= 2:
            v = d.eidx[key]
            ef = d.etype.faces
            fconec1 = f.conec
            fconec2 = d.conec[v][:, ef].reshape(-1, ef.shape[1])
            from fempy.domain.tools import find_faces
            fvalue = find_faces(fconec1, fconec2, sort=True)
            f.egroups[key] = fvalue  # recurrence
        d.eidx[key] = value     # TODO: recovers ngroups overwritten in faces

    def __delitem__(self, key):
        d = self.domain
        del d.eidx[key]
        f = d._faces
        if f is not None:
            del f.egroups[key]  # recurrence

    def __iter__(self):
        return iter(self.domain.eidx)

    def __len__(self):
        return len(self.domain.eidx)

    def update_global(self, *E, **F):
        if len(E) and E[0] is None:
            return
        EF = {}
        EF.update(*E, **F)
        for k, v in EF.iteritems():
            self.domain.eidx.g[k] = v  # set global eindex
            v = self.domain.egroups[k]     # get local
            self[k] = v  # set again, but face groups are also set now

    def set_group_by_conec(self, key, conec, glob=False):
        from fempy.domain.tools import find_faces
        d = self.domain
        if glob is False:
            dconec = d.eidx.conec
            vbool = find_faces(dconec, conec, sort=True)
            self[key] = vbool
        else:
            dconec = d.eidx.conec_g
            vbool = find_faces(dconec, conec, sort=True)
            self[key] = vbool
            # self.update_global({key: vbool})


class NGroups(collections.MutableMapping):
    def __init__(self, domain):
        self.domain = domain

    def __getitem__(self, key):
        if isinstance(key, str):
            if 'boundary' in key or 'BOUNDARY' in key or 'CUT' in key:
                self.domain.boundary  # creates boundary if not exists...
        return self.domain.eidx.nidx[key]

    def __setitem__(self, key, value):
        self.domain.eidx.nidx[key] = value

    def __delitem__(self, key):
        del self.domain.eidx.nidx[key]

    def __iter__(self):
        return iter(self.domain.eidx.nidx)

    def __len__(self):
        return len(self.domain.eidx.nidx)

    def update_global(self, *E, **F):
        if len(E) and E[0] is None:
            return
        EF = {}
        EF.update(*E, **F)
        for k, v in EF.iteritems():
            self.domain.eidx.nidx.g[k] = v  # set global eindex
            # v = self.domain.egroups[k]     # get local
            # self[k] = v  # set again, but face groups are also set now

    def nums(self, key=None, sortby=None, reverse=False):
        # Valid keys are: str and array-like of bool or int type
        # idx = np.nonzero(self[key])[0]
        if key is None:
            idx = np.arange(self.domain.nnode)
        elif isinstance(key, str):
            idx = np.nonzero(self[key])[0]
        else:
            key = np.asarray(key)
            if key.dtype == np.bool:
                idx = np.nonzero(self[key])[0]
            else:  # integer array
                idx = key
        if sortby is not None:
            coors = self.domain.coors[idx]
            for sc in sortby:
                c = 0 if sc in (0, 'x') else None
                c = 1 if c is None and sc in (1, 'y') else c
                c = 2 if c is None and sc in (2, 'z') else c
                assert c is not None, "Wrong sorting definition"
                assert c < coors.shape[1], "Sorry, cannot sort by {}-th "
                "dimension".format(c)
                si = coors[:, c].argsort()
                idx = idx[si]
                coors = coors[si]
        if reverse:
            idx = idx[::-1]
        return idx


if __name__ == '__main__':
    from fempy.testing import testmodule
    testmodule(__file__, '-v')



# -*- coding: utf-8 -*-
"""
Created on Sat Nov 22 22:47:46 2014

@author: mwojc
"""
import numpy as np
from fempy.domain.domain import Domain
from fempy.testing import testdata
from fempy.domain import elems


class Simple2D(object):
    """
      6 __7__ 8
       |\ |\ |
      3|_\4_\|5
       |\ |\ |
       |_\|_\|__
      0   1  2  9

    """
    coors = np.array([[0, 0],   # 0
                      [1, 0],   # 1
                      [2, 0],   # 2
                      [0, 1],   # 3
                      [1, 1],   # 4
                      [2, 1],   # 5
                      [0, 2],   # 6
                      [1, 2],   # 7
                      [2, 2],   # 8
                      [3, 0]],  # 9
                     dtype=float)
    conec2D = np.array([[0, 1, 3],   # 0
                        [1, 4, 3],   # 1
                        [1, 2, 4],   # 2
                        [2, 5, 4],   # 3
                        [3, 4, 6],   # 4
                        [4, 7, 6],   # 5
                        [4, 5, 7],   # 6
                        [5, 8, 7]],  # 7
                       dtype=int)
    conec1D = np.array([[0, 1],   # 0
                        [1, 2],   # 1
                        [2, 5],   # 2
                        [5, 8],   # 3
                        [8, 7],   # 4
                        [7, 6],   # 5
                        [6, 3],   # 6
                        [3, 0],   # 7
                        [2, 9]],  # 8
                       dtype=int)
    conec0D = np.array([[0],   # 0
                        [2],   # 1
                        [6],   # 2
                        [8],   # 3
                        [9]],  # 4
                       dtype=int)
    etype2D = elems.Tria31
    etype1D = elems.Line21
    etype0D = elems.Point10
    egroups2D = {'q1': [0, 1], 'q2': [2, 3], 'q3': [4, 5], 'q4': [6, 7]}
    egroups1D = {'b': [0, 1], 'r': [2, 3], 't': [4, 5], 'l': [6, 7], 'o': [8]}
    egroups0D = {'c1': [0], 'c2': [1], 'c3': [2], 'c4': [3], 'c5': [4]}
    ngroups = {'cline': [[3, 4], [4, 5]], 'cpoint': [4], 'opoint': [9],
               'cross': [6, 4, 2]}

    @property
    def domain2D(self):
        d = Domain(self.coors.copy(), self.conec2D, egroups=self.egroups2D,
                   ngroups=self.ngroups)
        f = d.faces
        ff = d.faces.faces
        for k, v in self.egroups1D.items():
            f.egroups.set_group_by_conec(k, self.conec1D[v], glob=True)
            # d.ngroups.update_global([(k, self.conec1D[v])])
        for k, v in self.egroups0D.items():
            ff.egroups.set_group_by_conec(k, self.conec0D[v], glob=True)
            # d.ngroups.update_global([(k, self.conec0D[v])])
        return d

    @property
    def domain1D(self):
        d = Domain(self.coors.copy(), self.conec1D, egroups=self.egroups1D,
                   ngroups=self.ngroups)
        f = d.faces
        for k, v in self.egroups0D.items():
            f.egroups.set_group_by_conec(k, self.conec0D[v], glob=True)
            # d.ngroups.update_global([(k, self.conec0D[v])])
        return d

    @property
    def domain0D(self):
        return Domain(self.coors.copy(), self.conec0D, egroups=self.egroups0D,
                      ngroups=self.ngroups)

    def domain2Dmv(self, v=[0, 0]):
        self.coors[4] += v  # move center point
        d = self.domain2D
        self.coors[4] -= v  # restore center point
        return d


class Simple3D(object):
    """
    Simple 3D domain
    """
    coors = np.loadtxt(testdata + 'simple3D.coors', dtype=float)

    conec3D = np.loadtxt(testdata + 'simple3D.conec3D', dtype=int)
    conec2D = np.loadtxt(testdata + 'simple3D.conec2D', dtype=int)
    conec1D = np.loadtxt(testdata + 'simple3D.conec1D', dtype=int)
    conec0D = np.loadtxt(testdata + 'simple3D.conec0D', dtype=int, ndmin=2)

    etype3D = elems.Tetr41
    etype2D = elems.Tria31
    etype1D = elems.Line21
    etype0D = elems.Point10

    egroups3D = {'center':  [0, 1, 2, 3],      # center elements
                 'bottome': [5, 10, 12, 6],    # elems with bottom faces
                 'lefte':   [5, 11, 9, 13],    # elems with left faces
                 'fronte':  [10, 13, 14, 15],  # elems with front faces
                 'righte':  [15, 12, 7, 8],    # elems with right faces
                 'backe':   [8, 6, 11, 4],     # elems with back faces
                 'tope':    [4, 9, 14, 7]}     # elems with top faces
    egroups2D = {'bottom':  [0, 1, 2, 3],      # z = 0
                 'left':    [4, 5, 6, 7],      # y = 0
                 'front':   [8, 9, 10, 11],    # x = 1
                 'right':   [12, 13, 14, 15],  # y = 1
                 'back':    [16, 17, 18, 19],  # x = 0
                 'top':     [20, 21, 22, 23]}  # z = 1
    egroups1D = {'bottoml': [0, 1, 2, 3],
                 'topl':    [4, 5, 6, 7],
                 'leftl':   [0, 4, 8, 9],
                 'frontl':  [1, 5, 9, 10],
                 'rightl':  [2, 6, 10, 11],
                 'backl':   [3, 7, 11, 8]}
    egroups0D = {'bottomc': [0, 1, 2, 3],
                 'topc':    [4, 5, 6, 7],
                 'leftc':   [0, 1, 4, 5],
                 'frontc':  [1, 2, 5, 6],
                 'rightc':  [2, 3, 6, 7],
                 'backc':   [3, 0, 7, 4]}
    ngroups = {'corners': [0, 1, 2, 3, 4, 5, 6, 7],
               'centers': [8, 9, 10, 11, 12, 13]}

    @property
    def domain3D(self):
        d = Domain(self.coors.copy(), self.conec3D, egroups=self.egroups3D,
                   ngroups=self.ngroups)
        f = d.faces
        ff = f.faces
        fff = ff.faces
        for k, v in self.egroups2D.items():
            f.egroups.set_group_by_conec(k, self.conec2D[v], glob=True)
        for k, v in self.egroups1D.items():
            ff.egroups.set_group_by_conec(k, self.conec1D[v], glob=True)
        for k, v in self.egroups0D.items():
            fff.egroups.set_group_by_conec(k, self.conec0D[v], glob=True)
        return d

    @property
    def domain2D(self):
        return Domain(self.coors, self.conec2D, egroups=self.egroups2D,
                      ngroups=self.ngroups)

    @property
    def domain1D(self):
        return Domain(self.coors, self.conec1D, egroups=self.egroups1D,
                      ngroups=self.ngroups)

    @property
    def domain0D(self):
        return Domain(self.coors, self.conec0D, egroups=self.egroups0D,
                      ngroups=self.ngroups)


def SquareWithCircle():
    """
    Produces square domain (a=1) with circular inclusion (r=0.25) for testing.
    """
    from fempy.geometry import Rectangle, Circle, Gmsh
    R = Rectangle(lb=(-0.5, -0.5))
    R.add_to_groups('r')
    C = Circle(r=0.25, center=(0, 0.))
    C.add_to_groups('c')
    R.add_hole(C)
    G = R + C  # now we have C in 'c, domain', R in 'r, domain'
    gm = Gmsh(G)
    return gm.domain()


if __name__ == "__main__":
    simple2D = Simple2D()
    simple2D.domain0D
    simple2D.domain1D
    simple2D.domain2D

    simple3D = Simple3D()
    simple3D.domain0D
    simple3D.domain1D
    simple3D.domain2D
    simple3D.domain3D

    SquareWithCircle()



# -*- coding: utf-8 -*-
import pytest
from fempy.geometry import *
import numpy as np


class TestPrimitivies:
    # Points
    A = Point([0, 0, 0], elsize=0.2, groups='A, corners')
    B = Point([1, 0, 0], elsize=0.1, groups=['B', 'corners'])
    C = Point([0, 1, 0], elsize=0.1, groups=['C', 'corners'])
    D = Point([0, -1e-8, 0], elsize=0.5, groups=['D', 'corners'])
    E = Point([2, 0, 0], elsize=0.5, groups=['E'])
    F = Point([1+np.sqrt(2), 0, 0], elsize=0.5, groups=['F'])
    G = Point([0.5, 0.5, 0], elsize=0.2, groups=['G', 'corners'])
    H = Point([0.25, 0.25, 0], elsize=0.2, groups=['H', 'corners'])
    I = Point([0, 0, 1e-6], groups=['I', 'corners'])  # out of plane
    J = Point([0, 0, 1], groups=['J', 'corners'])  # out of plane
    K = Point([0.0001, 0.9999, 0], groups=['K', 'corners'])
    L = Point([-2., 0, 0], groups=['L'])
    M = Point([0., -2., 0], groups=['M'])
    N = Point([0.5, 0.3, 0], groups=['N'])
    O = Point([0.5, 0.1, 0], groups=['O'])
    P = Point([0.5, -0.1, 0], groups=['P'])

#               C, ~K
#              o
#              |\
#              | \
#              |  \
#              |   o G
#              | H  \
#              |  o  \
#    o_________o______o______o___o
#    L         |A,D   B      E   F
#              |
#              |
#              o M

    # Lines
    AB = Line(A, B, groups=['AB', 'boundary'])
    BC = Line(B, C, groups=['BC', 'boundary'])
    CA = Line(C, A, groups=['CA', 'boundary'])

    BD = Line(B, D, groups=['BD', 'boundary'])
    DC = Line(D, C, groups=['DC', 'boundary'])
    CD = Line(C, D, groups=['CD', 'boundary'])
    AC = Line(A, C, groups=['AC', 'boundary'])
    CB = Line(C, B, groups=['CB', 'boundary'])
    AE = Line(A, E, groups=['AE', 'boundary'])
    BE = Line(B, E, groups=['BE', 'boundary'])
    CE = Line(C, E, groups=['CE', 'boundary'])
    CF = Line(C, F, groups=['CF', 'boundary'])
    CJ = Line(C, J, groups=['CJ'])
    JB = Line(J, B, groups=['JB'])
    AJ = Line(A, J, groups=['AJ'])
    AH = Line(A, H, groups=['AH'])
    HB = Line(H, B, groups=['HB'])
    AK = Line(A, K, groups=['AK'])
    KB = Line(K, B, groups=['KB'])
    IH = Line(I, H, groups=['IH'])
    IC = Line(I, C, groups=['IC'])
    AL = Line(A, L, groups=['AL'])
    LC = Line(L, C, groups=['LC'])

    # Splines
    ABCA = Spline([A, B, C, A], groups='closed, spline')
    ABCD = Spline([A, B, C, D], groups='closed, spline')
    ACBA = Spline([A, C, B, A], groups='closed, spline')
    ABC = Spline([A, B, C], groups='open, spline')
    ABEF = Spline([A, B, E, F], groups='open, spline')
    MFCLM = Spline([M, F, C, L, M], groups='boundary')
    ABGA = Spline([A, B, G, A], groups='hole')

    # Circular Arcs
    BAC = CircularArc(B, A, C, groups=['BAC'])
    BDC = CircularArc(B, D, C, groups=['BDC'])
    CAB = CircularArc(C, A, B, groups=['CAB'])
    CDB = CircularArc(C, D, B, groups=['CDB'])
    FBC = CircularArc(F, B, C, groups=['FBC'])
    BCJ = CircularArc(B, C, J, groups=['BCJ'])

    # Elliptical Arcs
    EAEC = EllipticalArc(E, A, E, C)
    EALC = EllipticalArc(E, A, L, C)
    CALE = EllipticalArc(C, A, L, E)
    CACE = EllipticalArc(C, A, C, E)
    ECEJ = EllipticalArc(E, C, E, J)
    EAFC = EllipticalArc(E, A, F, C)
    BALM = EllipticalArc(B, A, L, M)
    BAEM = EllipticalArc(B, A, E, M)
    BAEC = EllipticalArc(B, A, E, C)
    OAEM = EllipticalArc(O, A, E, M)
    OALM = EllipticalArc(O, A, E, M)
    OACL = EllipticalArc(O, A, C, L)
    MAMO = EllipticalArc(M, A, M, O)
    MALO = EllipticalArc(M, A, L, O)
    MAEO = EllipticalArc(M, A, E, O)

    # Plane surfaces
    ABCAs = PlaneSurface([ABCA], groups=['domain'])


class TestPoint(TestPrimitivies):
    def test_eq_ne(self):
        assert self.A != self.D
        assert self.A.eq(self.D)
        assert self.A.ne(self.B)

        class Pseudo:
            coors = np.array([0, 0, 0])
        P = Pseudo()
        assert not self.A.eq(P)

    def test_groups(self):
        assert self.A.groups == ['A', 'corners']
        self.A.remove_from_groups('corners')
        assert self.A.groups == ['A']
        self.A.add_to_groups('boundary')
        assert self.A.groups == ['A', 'boundary']


class TestLine(TestPrimitivies):
    def test_degenerated_line(self):
        with pytest.raises(AssertionError):
            Line(self.D, self.A)
        with pytest.raises(AssertionError):
            Line(self.A, self.D)

    def test_eq_ne(self):
        assert self.DC.eq(self.AC)
        assert self.DC.ne(self.AB)

    def test_is_reversed(self):
        assert self.DC.is_reversed(self.CD)
        assert self.DC.is_reversed(self.CA)

    def test_length(self):
        np.testing.assert_almost_equal(self.BC.length, np.sqrt(2))

    def test_poly(self):
        np.testing.assert_almost_equal(self.AB.poly(), self.AB.coors)
        #assert self.ABCD.poly(resolution=12).shape[0] == 12


class TestSpline(TestPrimitivies):
    def test_degenerated_line(self):
        with pytest.raises(AssertionError):
            Spline([self.D, self.A])

    def test_eq_ne(self):
        assert self.ABCA.eq(self.ABCD)

    def test_is_reversed(self):
        assert self.ABCD.is_reversed(self.ACBA)

    def test_is_closed(self):
        assert self.ABCD.is_closed is True
        assert self.ABCA.is_closed is True
        assert self.ABC.is_closed is False

    def test_poly(self):
        np.testing.assert_almost_equal(self.ABC.poly(), self.ABC.coors)
        assert self.ABCD.poly(res=12).shape[0] == 12
        x = [1, 0, -1, 0, 1]
        y = [0, 1, 0, -1, 0]
        pts = [Point(p) for p in zip(x, y)]
        S = Spline(pts)
        xy = S.poly(50)
        np.testing.assert_almost_equal(xy[0], xy[-1])
#        import pylab
#        pylab.plot(xy.T[0], xy.T[1])


class TestCircularArc(TestPrimitivies):
    def test_degenerated_arc(self):
        with pytest.raises(AssertionError):
            CircularArc(self.D, self.B, self.E)  # 180 degree
        with pytest.raises(AssertionError):
            CircularArc(self.A, self.C, self.B)  # not co-circular
        with pytest.raises(AssertionError):
            CircularArc(self.A, self.D, self.B)  # start eq center
        with pytest.raises(AssertionError):
            CircularArc(self.A, self.B, self.D)  # start eq end
        with pytest.raises(AssertionError):
            CircularArc(self.B, self.A, self.D)  # center eq end

    def test_eq_ne(self):
        assert self.BAC.eq(self.BDC)
        assert self.BAC.ne(self.FBC)

    def test_is_reversed(self):
        assert self.BAC.is_reversed(self.CDB)

    def test_radius(self):
        assert round(self.BAC.radius - 1., 7) == 0
        assert round(self.FBC.radius - np.sqrt(2), 7) == 0

    def test_angle(self):
        assert round(self.BAC.angle - np.pi/2., 7) == 0

    def test_chord(self):
        assert round(self.BAC.chord - np.sqrt(2), 7) == 0

    def test_length(self):
        assert round(self.BAC.length - 1*np.pi/2., 7) == 0

    def test_midpoint(self):
        target = np.array([0.57735027, -0.15470054,  0.57735027])
        check = abs(self.BCJ.midpoint.coors - target) < 1e-7
        assert check.all()

    def test_poly(self):
        np.testing.assert_almost_equal(self.BCJ.poly(), [self.BCJ.start.coors,
                                       self.BCJ.midpoint.coors,
                                       self.BCJ.end.coors])
        np.testing.assert_almost_equal(self.BCJ.poly(3)[1],
                                       self.BCJ.midpoint.coors)
        np.testing.assert_almost_equal(self.BCJ.poly(11)[[0, 5, 10]],
                                       [self.BCJ.start.coors,
                                        self.BCJ.midpoint.coors,
                                        self.BCJ.end.coors])


class TestEllipticalArc(TestPrimitivies):
    def test_degenerated_arc(self):
        with pytest.raises(AssertionError):
            EllipticalArc(self.D, self.B, self.D, self.E)  # colinear
        with pytest.raises(AssertionError):
            EllipticalArc(self.E, self.A, self.C, self.L)  # 180 degree
        with pytest.raises(ValueError):
            EllipticalArc(self.E, self.A, self.N, self.C)  # wrong radiuses

    def test_eq_ne(self):
        assert self.EAEC.eq(self.EAFC)
        assert self.EAEC.eq(self.EALC)
        assert self.MAMO.eq(self.MAEO)
        assert self.MAMO.eq(self.MALO)
        assert self.OAEM.eq(self.OALM)
        assert self.CALE.eq(self.CACE)
        assert self.OAEM.ne(self.MAMO)

    def test_is_reversed(self):
        assert self.OAEM.is_reversed(self.MAMO)
        assert self.EALC.is_reversed(self.CALE)

    def test_parametric_angle(self):
        spcangle = round(self.EALC.start_parametric_angle, 7)
        sprangle = round(self.EALC.start_polar_angle, 7)
        epcangle = round(self.EALC.end_parametric_angle, 7)
        eprangle = round(self.EALC.end_polar_angle, 7)
        assert spcangle == sprangle
        assert epcangle == eprangle

    def test_radiuses(self):
        aae = np.testing.assert_almost_equal
        aae(self.CALE.radiuses, (2, 1))
        aae(self.CACE.radiuses, (1, 2))

    def test_length(self):
        assert abs(self.EAEC.length - self.EALC.length) < 1e-7
        assert self.BAEC.length - np.pi/2. < 1e-7
        assert abs(self.EAEC.length - np.sqrt(93 + np.sqrt(3)/2)/4.) < 1e-7

    def test_midpoint(self):
        target = np.array([np.sqrt(2), np.sqrt(2)/2,  0.])
        check = abs(self.EAEC.midpoint.coors - target) < 1e-7
        assert check.all()

    def test_poly(self):
        aae = np.testing.assert_almost_equal
        aae(self.EAEC.poly(), [self.EAEC.start.coors,
                               self.EAEC.midpoint.coors,
                               self.EAEC.end.coors])
        aae(self.EAEC.poly(3)[1], self.EAEC.midpoint.coors)
        aae(self.EAEC.poly(11)[[0, 5, 10]], [self.EAEC.start.coors,
                                             self.EAEC.midpoint.coors,
                                             self.EAEC.end.coors])
        np.testing.assert_almost_equal(self.EAEC.poly(50), self.EALC.poly(50))
        x, y, z = self.EAEC.poly(50).T
        #import pylab
        #pylab.plot(x, y)
        #pylab.show()


class TestPlaneSurface(TestPrimitivies):
    def test_proper_definition(self):
        ABC = PlaneSurface([self.AB, self.BC, self.CA])
        assert ABC.curves == [self.AB, self.BC, self.CA]
        assert ABC.corient == [1, 1, 1]

    def test_repeating_lines(self):
        ABC = PlaneSurface([self.AB, self.BC, self.CA, self.CD, self.CA])
        assert ABC.curves == [self.AB, self.BC, self.CA]
        assert ABC.corient == [1, 1, 1]

    def test_not_connected_lines(self):
        ABC = PlaneSurface([self.AB, self.BC, self.CA, self.BE])
        assert ABC.curves == [self.AB, self.BC, self.CA]
        assert ABC.corient == [1, 1, 1]

    def test_overconnected_lines(self):
        with pytest.raises(ValueError):
            PlaneSurface([self.AB, self.BC, self.CA, self.BE, self.CE])

    def test_reorient_single_curve(self):
        ABC = PlaneSurface([self.AB, self.CB, self.CA])
        assert ABC.curves == [self.AB, self.CB, self.CA]
        assert ABC.corient == [1, -1, 1]

    def test_reorganize_curves(self):
        ABC = PlaneSurface([self.AB, self.CE, self.BE, self.AC])
        assert ABC.curves == [self.AB, self.BE, self.CE, self.AC]
        assert ABC.corient == [1, 1, -1, -1]

    def test_reorient_curve_loop(self):
        ABC = PlaneSurface([self.AB, self.CE, self.BE, self.AC],
                           orientation=-1)
        assert ABC.curves == [self.AB, self.AC, self.CE, self.BE]
        assert ABC.corient == [-1, 1, 1, -1]

    def test_cw_ccw_xy(self):
        ABC = PlaneSurface([self.DC, self.BD, self.BC], orientation='cw')
        assert ABC._PlaneSurface__get_orientation() == -1
        assert ABC.curves == [self.DC, self.BC, self.BD]
        assert ABC.corient == [1, -1, 1]
        ABC.orientation = 'counterclockwise'
        assert ABC.curves == [self.DC, self.BD, self.BC]
        assert ABC.corient == [-1, -1, 1]

    def test_cw_ccw_xz(self):
        ABC = PlaneSurface([self.JB, self.AB, self.AJ])
        assert ABC._PlaneSurface__get_orientation() == 1
        assert ABC.curves == [self.JB, self.AB, self.AJ]
        assert ABC.corient == [1, -1, 1]
        ABC.orientation = 'counterclockwise'
        assert ABC.curves == [self.JB, self.AB, self.AJ]
        assert ABC.corient == [1, -1, 1]

    def test_cw_ccw_yz(self):
        ABC = PlaneSurface([self.CA, self.CJ, self.AJ], orientation=-1)
        assert ABC._PlaneSurface__get_orientation() == -1
        assert ABC.curves == [self.CA, self.AJ, self.CJ]
        assert ABC.corient == [1, 1, -1]
        ABC.orientation = 'ccw'
        assert ABC.curves == [self.CA, self.CJ, self.AJ]
        assert ABC.corient == [-1, 1, -1]

    def test_cw_ccw_xyz(self):
        ABC = PlaneSurface([self.BC, self.CJ, self.JB])
        assert ABC._PlaneSurface__get_orientation() == 1
        assert ABC.curves == [self.BC, self.CJ, self.JB]
        assert ABC.corient == [1, 1, 1]
        ABC.orientation = 'cw'
        assert ABC.curves == [self.BC, self.JB, self.CJ]
        assert ABC.corient == [-1, -1, -1]

    def test_cw_ccw_negative(self):
        ABC = PlaneSurface([self.AL, self.LC, self.CA], orientation=-1)
        assert ABC._PlaneSurface__get_orientation() == -1
        ABC = PlaneSurface([self.AC, self.AL, self.LC])
        assert ABC._PlaneSurface__get_orientation() == 1

    def test_cw_ccw_concave(self):
        ABC = PlaneSurface([self.AH, self.HB, self.BC, self.CA])
        assert ABC.curves == [self.AH, self.HB, self.BC, self.CA]
        assert ABC.corient == [1, 1, 1, 1]

    def test_cw_ccw_very_concave(self):
        ABC = PlaneSurface([self.AK, self.KB, self.BC, self.CA],
                           orientation=-1)
        assert ABC.curves == [self.AK, self.CA, self.BC, self.KB]
        assert ABC.corient == [-1, -1, -1, -1]

    def test_cw_ccw_cieslik(self):
        A = Point([0, 0], groups=['corners'], elsize=1)
        B = Point([0, 462], groups=['corners'], elsize=1)
        B1 = Point([0, 457], groups=['corners'], elsize=1)
        C = Point([462, 462], groups=['corners'], elsize=1)
        D = Point([462, 0], groups=['corners'], elsize=1)
        D1 = Point([457, 0], groups=['corners'], elsize=1)

        B1B = Line(B1, B, groups=['left'])
        BC = Line(B, C, groups=['top'])
        CD = Line(C, D, groups=['right'])
        DD1 = Line(D, D1, groups=['bottom'])
        B1AD1 = CircularArc(B1, A, D1)

        BDC_Surface = PlaneSurface([B1B, BC, CD, DD1, B1AD1],
                                   groups=['matrix2'],
                                   elsize_factor=20.)

        assert BDC_Surface.curves == [B1B, B1AD1, DD1, CD, BC]
        assert BDC_Surface.corient == [-1, 1, -1, -1, -1]
        BDC_Surface.orientation = -1
        assert BDC_Surface.curves == [B1B, BC, CD, DD1, B1AD1]
        assert BDC_Surface.corient == [1, 1, 1, 1, -1]

    def test_non_coplanar(self):
        with pytest.raises(ValueError):
            PlaneSurface([self.AH, self.HB, self.JB, self.AJ])

    def test_almost_coplanar(self):
        with pytest.raises(ValueError):
            PlaneSurface([self.IH, self.HB, self.BC, self.IC])
        PlaneSurface([self.IH, self.HB, self.BC, self.IC], tol=1e-6)

    def test_non_coplanar_arc(self):
        with pytest.raises(ValueError):
            PlaneSurface([self.BAC, self.CJ, self.JB])

    def test_no_curve_loops(self):
        with pytest.raises(ValueError):
            PlaneSurface([self.BAC, self.CJ, self.AB, self.HB])

    def test_two_curves(self):
        ABC = PlaneSurface([self.BAC, self.BC])
        assert ABC.curves == [self.BAC, self.BC]
        assert ABC.corient == [1, -1]
        ABC = PlaneSurface([self.BC, self.BAC])
        assert ABC.curves == [self.BC, self.BAC]
        assert ABC.corient == [-1, 1]
        ABC = PlaneSurface([self.CAB, self.BC])
        assert ABC.curves == [self.CAB, self.BC]
        assert ABC.corient == [-1, -1]
        ABC = PlaneSurface([self.BC, self.CAB])
        assert ABC.curves == [self.BC, self.CAB]
        assert ABC.corient == [-1, -1]

    def test_two_curves_ellipse(self):
        ABC = PlaneSurface([self.EAEC, self.CE], 'two curves')
        ABC.gmsh.msh
        #ABC.gmsh.preview()

    def test_colinear_lines(self):
        with pytest.raises(ValueError):
            PlaneSurface([self.AB, self.BE, self.AE])

    def test_single_spline(self):
        ABCA = PlaneSurface([self.ABCA])
        ACBA = PlaneSurface([self.ACBA])
        assert ABCA.curves == [self.ABCA]
        assert ABCA.corient == [1]
        assert ACBA.corient == [-1]

    def test_gmsh_msh_with_hole(self):
        h = PlaneSurface([self.ABGA], groups='hole', remove=False)
        s = PlaneSurface([self.MFCLM], holes=[h], groups='domain')
        s.gmsh.msh
        # s.gmsh.preview()

    def test_surface_add(self):
        s1 = PlaneSurface([self.AB, self.BC, self.CA], groups='ABC')
        s2 = PlaneSurface([self.BE, self.CE, self.CB], groups='BEC')
        s = s1 + s2
        s.gmsh.msh
        #s.gmsh.preview()

    def test_surface_sub(self):
        s1 = PlaneSurface([self.AB, self.BC, self.CA], groups='ABC')
        s2 = PlaneSurface([self.BE, self.CE, self.CB], groups='BEC')
        s = s1 + s2
        s = s - s1
        s.gmsh.msh
        #s.gmsh.preview()

    def test_surface_move(self):
        s = PlaneSurface([self.AB, self.BC, self.CA], groups='ABC')
        s.move([2, 0])
        s.gmsh.msh
        #s.gmsh.preview()

    def test_surface_resize(self):
        s1 = PlaneSurface([self.AB, self.BC, self.CA], groups='ABC')
        s2 = s1.copy()
        s2.resize(0.5)
        s1.add_hole(s2)
        s1.gmsh.msh
        # s1.gmsh.preview()

    def test_periodic_lines(self):
        ef = self.A.elsize_factor
        self.A.elsize_factor=0.05
        self.AB.periodic(self.BC, -1)
        s1 = PlaneSurface([self.AB, self.BC, self.CA], groups='ABC')
        s1.gmsh.msh
        # s1.gmsh.preview()
        self.A.elsize_factor = ef
        self.AB.slaves = None  

if __name__ == '__main__':
    pytest.main([str(__file__), '-v'])  # Run tests from current file


# -*- coding: utf-8 -*-
"""
Created on Tue Nov 18 23:24:04 2014

@author: mwojc
"""
import numpy as np
from fempy.fields import tools
from fempy.domain.tools import cached_property, invalidate_caches
from fempy.fields.fields import Field


def transpose(a, *axes):
    '''
    Transposes only field axes
    '''
    if not isinstance(a, Field):
        return np.transpose(a, *axes)
    if a.sufix < 2:
        return a
    if len(axes) == 0:
        axes = tuple(reversed(a.saxes))
    axes = a.paxes + axes
    return np.transpose(a, axes)


def tensordot(a, b, axes=2):
    ap, af, bp, bf, rp = tools.get_prefixes(a, b)

    # If not fields make regular tensordot
    if ap == 0 and bp == 0:
        return np.tensordot(a, b, axes)

    # If there are fields use einsum
    pa = 'abcdefgh'[:ap]   # prefix
    pb = 'abcdefgh'[:bp]
    pr = 'abcdefgh'[:rp]
    sa = 'ijklmnop'[:af]   # sufix
    sb = 'qrstuvwx'[:bf]

    if isinstance(axes, int):
        axes = np.asarray([range(-axes, 0), range(0, axes)])
    else:
        axes = np.asarray(axes)
        if len(axes.shape) == 1:
            axes.shape = (2, 1)
    sar = sa
    sbr = sb
    for ai, bi in axes.T:
        sar = sar.replace(sa[ai], '')
        sbr = sbr.replace(sb[bi], '')
        sa = sa.replace(sa[ai], sb[bi])

    subscripts = '%s%s,%s%s->%s%s' % (pa, sa, pb, sb, pr, sar+sbr)
    r = np.einsum(subscripts, a, b)

    # Return type from field factory
    if hasattr(a, 'factory'):
        return a.factory(r)
    else:
        return b.factory(r)


def dot(a, b):
    # If not fields use regular np.dot
    if not isinstance(a, Field):
        if not isinstance(b, Field):
            return np.dot(a, b)
    # If fields use einsum (via tensordot)
    ap, af, bp, bf, rp = tools.get_prefixes(a, b)
    if af == 0 or bf == 0:  # scalar
        return tensordot(a, b, 0)
    if bf == 1:  # vector on the right
        return tensordot(a, b, 1)
    return tensordot(a, b, np.array([[-1], [-2]]))


def norm(a, ord=None, axis=None, keepdims=False):
    if not isinstance(a, Field):
        return np.linalg.norm(a, ord=ord, axis=axis, keepdims=keepdims)
    ax = a.saxes
    if axis is not None:
        ax = np.asarray(ax)[axis]
        if not isinstance(ax, (int, np.int, np.int64, np.int32, np.int8)):
            ax = tuple(ax)
    norm = np.linalg.norm(a, ord=ord, axis=ax, keepdims=keepdims)
    return a.factory(norm)


def normalize(a, ord=2, axis=-1):
    return a/norm(a, ord=ord, axis=axis, keepdims=True)


def flatten(a):
    return a.reshape(a.pshape + (np.prod(a.sshape, dtype=int),))


def identity_like(a):  # Valid for 2D fields
    n = a.sshape[-1]
    i = np.identity(n, dtype=a.dtype)
    return np.zeros_like(a) + i


def eye_like(a, k=0):  # Valid for 2D fields
    i = np.eye(*a.sshape, k=k, dtype=a.dtype)
    return np.zeros_like(a) + i


def psort(pvals, pvecs):
    M = np.prod(pvals.shape[:-1])
    N = pvals.shape[-1]
    ishape = pvals.shape[:-1] + (1,)
    #
    idx = np.argsort(-pvals)  # '-' is for descending order
    idx += np.arange(M).reshape(ishape)*N  # for sorting flattened array
    pvals = pvals.ravel()[idx]
    # pvecs = pvecs.reshape(M*N, -1)[idx]
    swap = np.swapaxes
    pvecs = swap(swap(pvecs, -1, -2).reshape(M*N, -1)[idx], -1, -2)
    return pvals, pvecs


def rotate(a, r):
    if isinstance(a, Field):
        r = a.factory(r)
        n = len(a.sshape)
    else:
        n = len(a.shape)
    if n == 0:  # scalar
        return a
    if n == 1:
        aprim = np.einsum('...ls,...s->...l', r, a)
    if n == 2:
        aprim = np.einsum('...kr,...ls,...rs->...kl',
                          r, r, a, optimize=True)
    if n == 3:
        aprim = np.einsum('...jq,...kr,...ls,...qrs->...jkl',
                          r, r, r, a, optimize=True)
    if n == 4:
        aprim = np.einsum('...ip,...jq,...kr,...ls,...pqrs->...ijkl',
                          r, r, r, r, a, optimize=True)
    if isinstance(a, Field):
        return a.factory(aprim)
    else:
        return a


class TensorCalc(object):
    def __init__(self, tensor=None):
        self.tensor = tensor

    @property
    def tensor(self):
        if not hasattr(self, '_tensor'):
            self._tensor = None
        if self._tensor is None:
            raise AttributeError("Set tensor first...")
        return self._tensor

    @tensor.setter
    def tensor(self, t):
        if t is not None:
            if not isinstance(t, Field):
                t = Field(t, len(t.shape) - 2)
            assert t.sufix == 2, 'Second rank tensor needed'
            assert t.shape[-2] == t.shape[-1], 'Second rank tensor needed'
            assert t.shape[-1] in (2, 3), '2D or 3D second rank tensor needed'
        self._tensor = t
        invalidate_caches(self)

    @cached_property
    def shape(self):
        return self.tensor.sshape

    @cached_property
    def T(self):
        return transpose(self.tensor)

    @cached_property
    def sym(self):
        return 0.5*(self.tensor + self.T)

    @cached_property
    def skew(self):
        return 0.5*(self.tensor - self.T)

    @cached_property
    def diag(self):
        t = self.tensor
        return np.diagonal(t, axis1=-2, axis2=-1)

    @cached_property
    def trace(self):
        t = self.tensor
        return np.trace(t, axis1=-2, axis2=-1)

    @cached_property
    def mean(self):
        trace = self.trace
        n = self.shape[-1]
        return trace/n

    @cached_property
    def dev(self):
        mean = self.mean
        dev = self.tensor.copy()
        n = self.shape[-1]
        for i in range(n):
            dev[..., i, i] -= mean
        return dev

    @cached_property
    def pvals(self):
        return self.eig[0]

    @cached_property
    def pvecs(self):
        return self.eig[1]

    @cached_property
    def kron(self):
        return eye_like(self.tensor)

    @cached_property
    def ptensor(self):
        pv = self.pvals  # possibly imaginary
        pt = np.zeros_like(self.tensor, dtype=pv.dtype)
        n = self.shape[-1]
        for i in range(n):
            pt[..., i, i] = pv[..., i]
        return pt

    @cached_property
    def polar(self):
        W, S, V = self.svd
        R = dot(W, V)
        VS = np.einsum('...ij,...i->...ij', V, S)  # S is diagonal
        U = dot(V, VS)
        return R, U

    @cached_property
    def svd(self):
        return np.linalg.svd(self.tensor)

    @cached_property
    def eig(self):
        return psort(*np.linalg.eig(self.tensor))

    @cached_property
    def inv(self):
        return np.linalg.inv(self.tensor)

    @cached_property
    def det(self):
        return np.linalg.det(self.tensor)

    @cached_property
    def I1(self):
        return self.trace

    @cached_property
    def I2(self):
        return 0.5*(self.trace**2 - self.tensordot(self.T))

    @cached_property
    def I3(self):
        return self.det

    @cached_property
    def _dev(self):
        return TensorCalc(self.dev)

    @cached_property
    def J2(self):
        return -self._dev.I2

    @cached_property
    def J3(self):
        return self._dev.I3

    @cached_property
    def sqrtJ2(self):
        return np.sqrt(self.J2)

    @cached_property
    def p(self):
        P = self.pvals
        return 0.5*(P[..., 0] + P[..., -1])

    @cached_property
    def q(self):
        P = self.pvals
        return 0.5*np.abs((P[..., 0] - P[..., -1]))

    @cached_property
    def b(self):
        P = self.pvals
        if self.shape[0] == 2:
            return 0.5
        return (P[..., 0] - P[..., 1])/(P[..., 0] - P[..., 2])

    @cached_property
    def a(self):
        return 2*self.b - 1.

    @cached_property
    def theta(self):
        return np.arctan(self.a/np.sqrt(3))

    def tensordot(self, other):
        return tensordot(self.tensor, other)

    def dot(self, other):
        return dot(self.tensor, other)


if __name__ == '__main__':
    from fempy.testing import testmodule
    testmodule(__file__, '-v')



# -*- coding: utf-8 -*-
"""
Created on Fri Apr  8 11:51:44 2016

@author: mwojc
"""
import scipy.sparse as sparse
import scipy.sparse.linalg
from fempy.solvers._solver import Solver

LOP = sparse.linalg.LinearOperator


class preconditioner(Solver):
    exact = False
    method = staticmethod(sparse.linalg.spilu)  # use this by default

    def inverse(self, A, **options):
        options = self.getopts(options)
        iA = self.method(A, **options)
        if hasattr(iA, 'solve'):
            iA = LOP(A.shape, dtype=A.dtype, matvec=iA.solve)
        else:
            iA = sparse.linalg.aslinearoperator(iA)
        return iA
    __call__ = inverse

    def solve(self, A, b, **options):
        x = self.inverse(A, **options).dot(b)
        self.info = 0
        return x


# INEXACT PRECONDITIONERS
class inexact(preconditioner):
    pass


class ilu(inexact):
    pass


class diag(inexact):
    def method(self, A, **options):
        iA = sparse.csr_matrix(A.shape, dtype=A.dtype)
        iA.setdiag(1./A.diagonal())
        return iA


# if pyamg is present
try:
    from pyamg import smoothed_aggregation_solver as sas

    class amg_sas(inexact):
        def method(self, A, **options):
            return sas(A, **options).aspreconditioner()
except ImportError:
    pass


# EXACT PRECONDITIONERS
class exact(preconditioner):
    exact = True
    method = staticmethod(sparse.linalg.splu)


class lu(exact):
    pass


# if umfpack is present
try:
    from scikits import umfpack as umf

    class umfpack(exact):
        exact = True

        def method(self, A, **options):
            if hasattr(A, 'has_sorted_indices'):
                umf.assumeSortedIndices = A.has_sorted_indices
            return umf.splu(A, **options)
except ImportError:
    pass


# if pyamgcl is present
try:
    import pyamgcl

    class amgcl_cg(exact):  # We create pyamg solver actually here
        def method(self, A, **options):
            options = self.getopts(options)
            options['type'] = 'cg'
            options['maxiter'] = 1000
            if not sparse.issparse(A):
                A = sparse.csr_matrix(A)
            S = pyamgcl.solver(pyamgcl.amg(A), prm=options)

            def matvec(b):
                x = S(b.ravel())
                return x.reshape(b.shape)  # necesacry for matmat
            #return LOP(A.shape, dtype=A.dtype, matvec=matvec)
            return pyamgcl.amg(A)
except ImportError:
    pass


# lu_default_opts = {'perm_spec': None,
#                    'diag_pivot_tresh': None,
#                    'drop_tol': None,
#                    'relax': None,
#                    'panel_size': None,
#                    'options': {}}
# ilu_default_opts = {'drop_tol': None,
#                    'fill_factor': None,
#                    'drop_rule': None,
#                    'perm_spec': None,
#                    'diag_pivot_tresh': None,
#                    'drop_tol': None,
#                    'relax': None,
#                    'panel_size': None,
#                    'options': {}}


if __name__ == '__main__':
    from fempy.testing import testmodule
    testmodule(__file__, '-v')
